from __future__ import annotations

import asyncio
import logging
from contextlib import asynccontextmanager
from importlib import import_module
from pathlib import Path
from typing import Any, cast

from fastapi import FastAPI, HTTPException
from fastapi.responses import FileResponse

from app import __version__
from app.core.clients.http import close_http_client, init_http_client
from app.core.config.settings import get_settings
from app.core.config.settings_cache import get_settings_cache
from app.core.handlers import add_exception_handlers
from app.core.metrics.middleware import MetricsMiddleware
from app.core.metrics.prometheus import MULTIPROCESS_MODE, PROMETHEUS_AVAILABLE, make_scrape_registry, mark_process_dead
from app.core.middleware import (
    add_api_firewall_middleware,
    add_request_decompression_middleware,
    add_request_id_middleware,
)
from app.core.middleware.inflight import InFlightMiddleware
from app.core.openai.model_refresh_scheduler import build_model_refresh_scheduler
from app.core.resilience.backpressure import BackpressureMiddleware
from app.core.resilience.bulkhead import BulkheadMiddleware, get_bulkhead
from app.core.resilience.memory_monitor import configure as configure_memory_monitor
from app.core.usage.refresh_scheduler import build_usage_refresh_scheduler
from app.db.session import SessionLocal, close_db, init_background_db, init_db
from app.modules.accounts import api as accounts_api
from app.modules.api_keys import api as api_keys_api
from app.modules.audit import api as audit_api
from app.modules.dashboard import api as dashboard_api
from app.modules.dashboard_auth import api as dashboard_auth_api
from app.modules.firewall import api as firewall_api
from app.modules.health import api as health_api
from app.modules.oauth import api as oauth_api
from app.modules.proxy import api as proxy_api
from app.modules.proxy.rate_limit_cache import get_rate_limit_headers_cache
from app.modules.proxy.ring_membership import (
    RING_HEARTBEAT_INTERVAL_SECONDS,
    RING_STALE_GRACE_SECONDS,
    RING_STALE_THRESHOLD_SECONDS,
    RingMembershipService,
    dynamic_bridge_ring_membership_enabled,
)
from app.modules.request_logs import api as request_logs_api
from app.modules.settings import api as settings_api
from app.modules.sticky_sessions import api as sticky_sessions_api
from app.modules.sticky_sessions.cleanup_scheduler import build_sticky_session_cleanup_scheduler
from app.modules.usage import api as usage_api
from app.modules.usage.additional_quota_keys import reload_additional_quota_registry

logger = logging.getLogger(__name__)


def _is_benign_metrics_bind_failure(exc: BaseException) -> bool:
    if not MULTIPROCESS_MODE:
        return False
    if isinstance(exc, SystemExit):
        return exc.code == 1
    if isinstance(exc, OSError):
        import errno as _errno

        return exc.errno in (_errno.EADDRINUSE, _errno.EADDRNOTAVAIL)
    return False


@asynccontextmanager
async def lifespan(app: FastAPI):
    import app.core.startup as startup_module

    shutdown_state = import_module("app.core.shutdown")
    metrics_server = None
    metrics_server_task: asyncio.Task[None] | None = None
    ring_service = None
    heartbeat_task: asyncio.Task[None] | None = None
    instance_id = None

    startup_module._startup_complete = False
    shutdown_state.reset()
    await get_settings_cache().invalidate()
    await get_rate_limit_headers_cache().invalidate()
    reload_additional_quota_registry()
    settings = get_settings()
    if settings.otel_enabled:
        from app.core.tracing.otel import init_tracing

        init_tracing(service_name="codex-lb", endpoint=settings.otel_exporter_endpoint, app=app)
    await init_db()
    init_background_db()
    await init_http_client()
    usage_scheduler = build_usage_refresh_scheduler()
    model_scheduler = build_model_refresh_scheduler()
    sticky_session_cleanup_scheduler = build_sticky_session_cleanup_scheduler()
    await usage_scheduler.start()
    await model_scheduler.start()
    await sticky_session_cleanup_scheduler.start()
    if settings.metrics_enabled and PROMETHEUS_AVAILABLE:
        import uvicorn

        scrape_registry = make_scrape_registry()
        prometheus_module = import_module("prometheus_client")
        make_asgi_app = getattr(prometheus_module, "make_asgi_app")
        metrics_app = make_asgi_app(registry=scrape_registry)
        config = uvicorn.Config(metrics_app, host="0.0.0.0", port=settings.metrics_port, log_level="warning")
        metrics_server = uvicorn.Server(config)

        async def _serve_metrics(srv: Any) -> None:
            try:
                await srv.serve()
            except SystemExit as exc:
                if _is_benign_metrics_bind_failure(exc):
                    logger.info(
                        "Metrics port %d unavailable (another worker likely serves metrics)",
                        settings.metrics_port,
                    )
                else:
                    raise
            except OSError as exc:
                if _is_benign_metrics_bind_failure(exc):
                    logger.info(
                        "Metrics port %d already bound (another worker serves metrics)",
                        settings.metrics_port,
                    )
                else:
                    raise

        metrics_server_task = asyncio.create_task(_serve_metrics(metrics_server))
    elif settings.metrics_enabled:
        logger.warning("Metrics endpoint enabled but prometheus-client is not installed")

    async def _register_and_heartbeat(svc: RingMembershipService, iid: str) -> None:
        attempt = 0
        while True:
            attempt += 1
            try:
                await svc.register(iid)
                logger.info("Registered in bridge ring", extra={"instance_id": iid, "attempt": attempt})
                break
            except Exception:
                delay = min(5.0 * (2 ** min(attempt - 1, 5)), 60.0)
                logger.warning("Ring registration attempt %d failed, retrying in %.0fs", attempt, delay, exc_info=True)
                await asyncio.sleep(delay)
        while True:
            await asyncio.sleep(RING_HEARTBEAT_INTERVAL_SECONDS)
            try:
                await svc.heartbeat(iid)
            except Exception:
                logger.warning("Ring heartbeat failed", exc_info=True)

    async def _heartbeat_only(svc: RingMembershipService, iid: str) -> None:
        while True:
            await asyncio.sleep(RING_HEARTBEAT_INTERVAL_SECONDS)
            try:
                await svc.heartbeat(iid)
            except Exception:
                logger.warning("Ring heartbeat failed", exc_info=True)

    ring_service: RingMembershipService | None = None
    instance_id: str | None = None
    heartbeat_task: asyncio.Task[None] | None = None
    if dynamic_bridge_ring_membership_enabled(settings):
        try:
            ring_service = RingMembershipService(SessionLocal)
            instance_id = settings.http_responses_session_bridge_instance_id
            await asyncio.wait_for(ring_service.register(instance_id), timeout=5.0)
            logger.info("Registered in bridge ring", extra={"instance_id": instance_id})
            heartbeat_task = asyncio.create_task(_heartbeat_only(ring_service, instance_id))
        except Exception:
            logger.warning("Ring registration failed, retrying in background", exc_info=True)
            if ring_service is not None and instance_id is not None:
                heartbeat_task = asyncio.create_task(_register_and_heartbeat(ring_service, instance_id))

    from app.core.auth.api_key_cache import get_api_key_cache
    from app.core.cache.invalidation import (
        NAMESPACE_API_KEY,
        NAMESPACE_FIREWALL,
        CacheInvalidationPoller,
        set_cache_invalidation_poller,
    )
    from app.core.middleware.firewall_cache import get_firewall_ip_cache

    cache_poller = CacheInvalidationPoller(SessionLocal)
    cache_poller.on_invalidation(NAMESPACE_API_KEY, get_api_key_cache().clear)
    cache_poller.on_invalidation(NAMESPACE_FIREWALL, get_firewall_ip_cache().invalidate_all)
    set_cache_invalidation_poller(cache_poller)
    await cache_poller.start()

    startup_module._startup_complete = True

    try:
        yield
    finally:
        shutdown_state.set_bridge_drain_active(True)
        shutdown_state.set_draining(True)
        drained = await shutdown_state.wait_for_in_flight_drain(timeout_seconds=settings.shutdown_drain_timeout_seconds)
        if not drained:
            logger.warning("Drain timeout reached, proceeding with shutdown")

        proxy_service = getattr(app.state, "proxy_service", None)
        if proxy_service is not None and hasattr(proxy_service, "close_all_http_bridge_sessions"):
            try:
                await proxy_service.close_all_http_bridge_sessions()
            except Exception:
                logger.warning("Failed to close HTTP bridge sessions during shutdown", exc_info=True)

        # Cancel heartbeat and age the shared ring row near expiry.
        if heartbeat_task is not None:
            heartbeat_task.cancel()
            try:
                await asyncio.wait_for(heartbeat_task, timeout=2)
            except (asyncio.CancelledError, TimeoutError):
                pass

        if ring_service is not None and instance_id is not None:
            try:
                await asyncio.wait_for(
                    ring_service.mark_stale(
                        instance_id,
                        stale_threshold_seconds=RING_STALE_THRESHOLD_SECONDS,
                        grace_seconds=RING_STALE_GRACE_SECONDS,
                    ),
                    timeout=3,
                )
                logger.info(
                    "Marked bridge ring membership stale for shutdown",
                    extra={"instance_id": instance_id},
                )
            except Exception:
                logger.warning("Failed to mark bridge ring membership stale during shutdown", exc_info=True)

        if metrics_server is not None:
            metrics_server.should_exit = True

        await cache_poller.stop()
        await sticky_session_cleanup_scheduler.stop()
        await model_scheduler.stop()
        await usage_scheduler.stop()
        try:
            await close_http_client()
        finally:
            try:
                if metrics_server_task is not None:
                    await asyncio.wait_for(metrics_server_task, timeout=5)
            except TimeoutError:
                logger.warning("Timed out waiting for metrics server shutdown")
            except Exception:
                logger.exception("Metrics server stopped with an error")
            finally:
                shutdown_state.reset()
                mark_process_dead()
                await close_db()


def create_app() -> FastAPI:
    settings = get_settings()
    configure_memory_monitor(
        warning_threshold_mb=settings.memory_warning_threshold_mb,
        reject_threshold_mb=settings.memory_reject_threshold_mb,
    )
    app = FastAPI(
        title="codex-lb",
        version=__version__,
        lifespan=lifespan,
        swagger_ui_parameters={"persistAuthorization": True},
    )

    app.add_middleware(cast(Any, InFlightMiddleware))
    add_request_decompression_middleware(app)
    add_request_id_middleware(app)
    add_api_firewall_middleware(app)
    app.add_middleware(cast(Any, MetricsMiddleware), enabled=settings.metrics_enabled)
    if settings.backpressure_max_concurrent_requests > 0:
        app.add_middleware(
            cast(Any, BackpressureMiddleware),
            max_concurrent=settings.backpressure_max_concurrent_requests,
        )
    app.add_middleware(
        cast(Any, BulkheadMiddleware),
        bulkhead=get_bulkhead(
            proxy_limit=settings.bulkhead_proxy_limit,
            dashboard_limit=settings.bulkhead_dashboard_limit,
        ),
    )
    add_exception_handlers(app)

    app.include_router(proxy_api.router)
    app.include_router(proxy_api.ws_router)
    app.include_router(proxy_api.v1_router)
    app.include_router(proxy_api.v1_ws_router)
    app.include_router(proxy_api.transcribe_router)
    app.include_router(proxy_api.usage_router)
    app.include_router(audit_api.router)
    app.include_router(accounts_api.router)
    app.include_router(dashboard_api.router)
    app.include_router(usage_api.router)
    app.include_router(request_logs_api.router)
    app.include_router(oauth_api.router)
    app.include_router(dashboard_auth_api.router)
    app.include_router(settings_api.router)
    app.include_router(firewall_api.router)
    app.include_router(sticky_sessions_api.router)
    app.include_router(api_keys_api.router)
    app.include_router(health_api.router)

    static_dir = Path(__file__).parent / "static"
    index_html = static_dir / "index.html"
    static_root = static_dir.resolve()
    frontend_build_hint = "Frontend assets are missing. Run `cd frontend && bun run build`."
    excluded_prefixes = ("api/", "v1/", "backend-api/", "health")

    def _is_static_asset_path(path: str) -> bool:
        if path.startswith("assets/"):
            return True
        last_segment = path.rsplit("/", maxsplit=1)[-1]
        return "." in last_segment

    @app.get("/", include_in_schema=False)
    @app.get("/{path:path}", include_in_schema=False)
    async def spa_fallback(path: str = ""):
        normalized = path.lstrip("/")
        if normalized and any(
            normalized == prefix.rstrip("/") or normalized.startswith(prefix) for prefix in excluded_prefixes
        ):
            raise HTTPException(status_code=404, detail="Not Found")

        if normalized:
            candidate = (static_dir / normalized).resolve()
            if candidate.is_relative_to(static_root) and candidate.is_file():
                return FileResponse(candidate)
            if _is_static_asset_path(normalized):
                raise HTTPException(status_code=404, detail="Not Found")

        if not index_html.is_file():
            raise HTTPException(status_code=503, detail=frontend_build_hint)

        return FileResponse(index_html, media_type="text/html")

    return app


app = create_app()



from __future__ import annotations

import logging
import uuid
from datetime import UTC, datetime, timedelta

from sqlalchemy import select, text

from app.core.config.settings import get_settings
from app.db.models import SchedulerLeader
from app.db.session import get_session

logger = logging.getLogger(__name__)


class LeaderElection:
    def __init__(self, leader_id: str | None = None) -> None:
        self._leader_id = leader_id or str(uuid.uuid4())
        self._is_leader = False

    async def try_acquire(self) -> bool:
        settings = get_settings()
        if not settings.leader_election_enabled:
            self._is_leader = True
            return True

        if "sqlite" in settings.database_url.lower():
            self._is_leader = True
            return True

        ttl = settings.leader_election_ttl_seconds
        now = datetime.now(UTC)
        expires_at = now + timedelta(seconds=ttl)

        try:
            async for session in get_session():
                await session.execute(
                    text(
                        """
                        INSERT INTO scheduler_leader (id, leader_id, acquired_at, expires_at)
                        VALUES (1, :leader_id, :now, :expires_at)
                        ON CONFLICT (id) DO UPDATE SET
                            leader_id = :leader_id,
                            acquired_at = :now,
                            expires_at = :expires_at
                        WHERE scheduler_leader.expires_at < :now OR scheduler_leader.leader_id = :leader_id
                        """
                    ),
                    {
                        "leader_id": self._leader_id,
                        "now": now,
                        "expires_at": expires_at,
                    },
                )
                await session.commit()
                row = await session.scalar(select(SchedulerLeader.leader_id).where(SchedulerLeader.id == 1))
                self._is_leader = row == self._leader_id
                return self._is_leader
        except Exception:
            logger.warning("Leader election failed, defaulting to non-leader", exc_info=True)

        self._is_leader = False
        return False

    async def renew(self) -> bool:
        if not self._is_leader:
            return False

        settings = get_settings()
        if not settings.leader_election_enabled:
            return True
        if "sqlite" in settings.database_url.lower():
            return True

        ttl = settings.leader_election_ttl_seconds
        expires_at = datetime.now(UTC) + timedelta(seconds=ttl)
        try:
            async for session in get_session():
                await session.execute(
                    text(
                        "UPDATE scheduler_leader SET expires_at = :expires_at WHERE id = 1 AND leader_id = :leader_id"
                    ),
                    {
                        "expires_at": expires_at,
                        "leader_id": self._leader_id,
                    },
                )
                await session.commit()
                return True
        except Exception:
            logger.warning("Failed to renew leadership", exc_info=True)
            return False

        return False


_leader_election: LeaderElection | None = None


def get_leader_election() -> LeaderElection:
    global _leader_election
    if _leader_election is None:
        _leader_election = LeaderElection()
    return _leader_election



from __future__ import annotations

import argparse
import logging
import sqlite3
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Sequence

from app.core.config.settings import get_settings
from app.db.sqlite_utils import IntegrityCheck, check_sqlite_integrity, sqlite_db_path_from_url

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class RecoveryOptions:
    source: Path
    output: Path
    replace: bool


@dataclass(slots=True)
class RecoveryOutcome:
    source: Path
    output: Path
    replaced: bool
    integrity: IntegrityCheck


def _timestamp() -> str:
    return datetime.now(timezone.utc).strftime("%Y%m%d-%H%M%S")


def _default_output_path(source: Path) -> Path:
    suffix = source.suffix or ".db"
    return source.with_name(f"{source.stem}.recover-{_timestamp()}{suffix}")


def _load_dump(source: Path) -> str:
    try:
        with sqlite3.connect(str(source)) as conn:
            return "\n".join(conn.iterdump())
    except sqlite3.DatabaseError as exc:
        message = f"failed to read sqlite dump: {exc}"
        raise RuntimeError(message) from exc


def _write_dump(output: Path, dump: str) -> None:
    try:
        with sqlite3.connect(str(output)) as conn:
            conn.execute("PRAGMA foreign_keys=OFF")
            conn.executescript(dump)
            conn.execute("PRAGMA foreign_keys=ON")
    except sqlite3.DatabaseError as exc:
        message = f"failed to write sqlite dump: {exc}"
        raise RuntimeError(message) from exc


def recover_sqlite_db(options: RecoveryOptions) -> RecoveryOutcome:
    if not options.source.exists():
        raise FileNotFoundError(f"sqlite database not found: {options.source}")
    if options.output.exists():
        raise FileExistsError(f"output database already exists: {options.output}")

    integrity = check_sqlite_integrity(options.source)
    if not integrity.ok:
        logger.warning("SQLite integrity check failed details=%s", integrity.details)
    else:
        logger.info("SQLite integrity check OK. Proceeding with export/import.")

    dump = _load_dump(options.source)
    _write_dump(options.output, dump)

    if options.replace:
        backup = options.source.with_name(f"{options.source.name}.corrupt-{_timestamp()}")
        options.source.replace(backup)
        options.output.replace(options.source)
        return RecoveryOutcome(
            source=backup,
            output=options.source,
            replaced=True,
            integrity=integrity,
        )

    return RecoveryOutcome(
        source=options.source,
        output=options.output,
        replaced=False,
        integrity=integrity,
    )


def _parse_args(argv: Sequence[str] | None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Recover a sqlite database via .dump/.executescript.")
    parser.add_argument("--db", help="Path to sqlite database (defaults to settings.database_url)")
    parser.add_argument(
        "--output",
        help="Output sqlite database path (default: <db>.recover-<timestamp>.db)",
    )
    parser.add_argument(
        "--replace",
        action="store_true",
        help="Replace the source database after recovery (source is renamed with .corrupt-<timestamp>)",
    )
    return parser.parse_args(args=argv)


def _resolve_source_path(db_path: str | None) -> Path:
    if db_path:
        return Path(db_path).expanduser()
    settings = get_settings()
    resolved = sqlite_db_path_from_url(settings.database_url)
    if resolved is None:
        raise RuntimeError("database_url is not a sqlite file path")
    return resolved


def main(argv: Sequence[str] | None = None) -> int:
    args = _parse_args(argv)
    source = _resolve_source_path(args.db)
    output = Path(args.output).expanduser() if args.output else _default_output_path(source)
    outcome = recover_sqlite_db(RecoveryOptions(source=source, output=output, replace=bool(args.replace)))
    if outcome.replaced:
        logger.info("Recovered database written to %s (original saved at %s)", outcome.output, outcome.source)
    else:
        logger.info("Recovered database written to %s", outcome.output)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from datetime import timedelta

import pytest

from app.core.crypto import TokenEncryptor
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus
from app.db.session import SessionLocal
from app.modules.accounts.repository import AccountsRepository
from app.modules.usage.repository import UsageRepository

pytestmark = pytest.mark.integration


def _make_account(account_id: str, email: str, plan_type: str = "plus") -> Account:
    encryptor = TokenEncryptor()
    return Account(
        id=account_id,
        email=email,
        plan_type=plan_type,
        access_token_encrypted=encryptor.encrypt("access"),
        refresh_token_encrypted=encryptor.encrypt("refresh"),
        id_token_encrypted=encryptor.encrypt("id"),
        last_refresh=utcnow(),
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )


@pytest.mark.asyncio
async def test_usage_summary_empty_returns_zeroes(async_client):
    response = await async_client.get("/api/usage/summary")
    assert response.status_code == 200
    payload = response.json()

    primary = payload["primaryWindow"]
    assert primary["remainingPercent"] == 0.0
    assert primary["capacityCredits"] == 0.0
    assert primary["remainingCredits"] == 0.0
    assert primary["windowMinutes"] == 300

    secondary = payload["secondaryWindow"]
    assert secondary["remainingPercent"] == 0.0
    assert secondary["capacityCredits"] == 0.0
    assert secondary["remainingCredits"] == 0.0
    assert secondary["windowMinutes"] == 10080

    cost = payload["cost"]
    assert cost["currency"] == "USD"
    assert cost["totalUsd7d"] == 0.0

    metrics = payload["metrics"]
    assert metrics["requests7d"] == 0
    assert metrics["tokensSecondaryWindow"] == 0
    assert metrics["cachedTokensSecondaryWindow"] == 0
    assert metrics["errorRate7d"] is None
    assert metrics["topError"] is None


@pytest.mark.asyncio
async def test_usage_history_aggregates_per_account(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)

        await accounts_repo.upsert(_make_account("acc_a", "a@example.com"))
        await accounts_repo.upsert(_make_account("acc_b", "b@example.com"))

        await usage_repo.add_entry("acc_a", 10.0, recorded_at=now - timedelta(hours=3))
        await usage_repo.add_entry("acc_a", 30.0, recorded_at=now - timedelta(hours=2))

    response = await async_client.get("/api/usage/history?hours=24")
    assert response.status_code == 200
    payload = response.json()
    assert payload["windowHours"] == 24

    accounts = {item["accountId"]: item for item in payload["accounts"]}
    acc_a = accounts["acc_a"]
    acc_b = accounts["acc_b"]

    assert acc_a["remainingPercentAvg"] == pytest.approx(80.0)
    assert acc_a["capacityCredits"] == pytest.approx(225.0)
    assert acc_a["remainingCredits"] == pytest.approx(180.0)

    assert acc_b["remainingPercentAvg"] == pytest.approx(100.0)
    assert acc_b["capacityCredits"] == pytest.approx(225.0)
    assert acc_b["remainingCredits"] == pytest.approx(225.0)


@pytest.mark.asyncio
async def test_usage_window_secondary_uses_latest_window_minutes(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)

        await accounts_repo.upsert(_make_account("acc_sec", "sec@example.com"))
        await usage_repo.add_entry(
            "acc_sec",
            40.0,
            window="secondary",
            reset_at=1735689600,
            window_minutes=1440,
            recorded_at=now - timedelta(minutes=5),
        )
    response = await async_client.get("/api/usage/window?window=secondary")
    assert response.status_code == 200
    payload = response.json()
    assert payload["windowKey"] == "secondary"
    assert payload["windowMinutes"] == 1440

    accounts = {item["accountId"]: item for item in payload["accounts"]}
    entry = accounts["acc_sec"]
    assert entry["remainingPercentAvg"] == pytest.approx(60.0)
    assert entry["capacityCredits"] == pytest.approx(7560.0)
    assert entry["remainingCredits"] == pytest.approx(4536.0)


@pytest.mark.asyncio
async def test_usage_window_primary_excludes_weekly_only_accounts(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)

        await accounts_repo.upsert(_make_account("acc_plus", "plus@example.com", plan_type="plus"))
        await accounts_repo.upsert(_make_account("acc_free", "free@example.com", plan_type="free"))
        await usage_repo.add_entry(
            "acc_plus",
            20.0,
            window="primary",
            window_minutes=300,
            recorded_at=now - timedelta(minutes=2),
        )
        await usage_repo.add_entry(
            "acc_free",
            20.0,
            window="primary",
            window_minutes=10080,
            recorded_at=now - timedelta(minutes=1),
        )

    response = await async_client.get("/api/usage/window?window=primary")
    assert response.status_code == 200
    payload = response.json()
    assert payload["windowKey"] == "primary"
    assert payload["windowMinutes"] == 300

    accounts = {item["accountId"]: item for item in payload["accounts"]}
    assert accounts["acc_free"]["remainingPercentAvg"] is None


@pytest.mark.asyncio
async def test_usage_window_secondary_includes_weekly_only_primary_entries(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)

        await accounts_repo.upsert(_make_account("acc_with_secondary", "with-secondary@example.com", plan_type="plus"))
        await accounts_repo.upsert(_make_account("acc_weekly_only", "weekly-only@example.com", plan_type="free"))

        await usage_repo.add_entry(
            "acc_with_secondary",
            40.0,
            window="secondary",
            window_minutes=10080,
            recorded_at=now - timedelta(minutes=1),
        )
        await usage_repo.add_entry(
            "acc_weekly_only",
            60.0,
            window="primary",
            window_minutes=10080,
            recorded_at=now - timedelta(minutes=1),
        )

    response = await async_client.get("/api/usage/window?window=secondary")
    assert response.status_code == 200
    payload = response.json()
    accounts = {item["accountId"]: item for item in payload["accounts"]}

    assert accounts["acc_with_secondary"]["remainingPercentAvg"] == pytest.approx(60.0)
    assert accounts["acc_weekly_only"]["remainingPercentAvg"] == pytest.approx(40.0)


@pytest.mark.asyncio
async def test_usage_history_team_plan_has_capacity(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)

        await accounts_repo.upsert(_make_account("acc_team", "team@example.com", plan_type="team"))
        await usage_repo.add_entry(
            "acc_team",
            20.0,
            window="primary",
            recorded_at=now - timedelta(hours=1),
        )

    response = await async_client.get("/api/usage/history?hours=24")
    assert response.status_code == 200
    payload = response.json()
    accounts = {item["accountId"]: item for item in payload["accounts"]}
    entry = accounts["acc_team"]

    assert entry["remainingPercentAvg"] == pytest.approx(80.0)
    assert entry["capacityCredits"] == pytest.approx(225.0)
    assert entry["remainingCredits"] == pytest.approx(180.0)


@pytest.mark.asyncio
async def test_usage_history_invalid_hours_returns_validation_error(async_client):
    response = await async_client.get("/api/usage/history?hours=0")
    assert response.status_code == 422
    payload = response.json()
    assert payload["error"]["code"] == "validation_error"


@pytest.mark.asyncio
async def test_usage_window_invalid_query_returns_validation_error(async_client):
    response = await async_client.get("/api/usage/window?window=invalid")
    assert response.status_code == 422
    payload = response.json()
    assert payload["error"]["code"] == "validation_error"



"""merge request-log transport heads

Revision ID: 20260313_024500_merge_request_log_transport_heads
Revises:
    20260312_020000_merge_request_logs_transport_and_feature_heads,
    20260312_020000_merge_request_logs_transport_and_sticky_heads
Create Date: 2026-03-13 02:45:00.000000
"""

from __future__ import annotations

# revision identifiers, used by Alembic.
revision = "20260313_024500_merge_request_log_transport_heads"
down_revision = (
    "20260312_020000_merge_request_logs_transport_and_feature_heads",
    "20260312_020000_merge_request_logs_transport_and_sticky_heads",
)
branch_labels = None
depends_on = None


def upgrade() -> None:
    return


def downgrade() -> None:
    return



from __future__ import annotations

import builtins
import importlib
import sys
import types
from collections.abc import Iterator

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient

pytestmark = pytest.mark.unit


class _MetricChild:
    def __init__(self) -> None:
        self.value = 0.0
        self.observations: list[float] = []

    def inc(self, amount: float = 1.0) -> None:
        self.value += amount

    def dec(self, amount: float = 1.0) -> None:
        self.value -= amount

    def observe(self, amount: float) -> None:
        self.observations.append(amount)


class _MetricBase:
    def __init__(self, name: str, documentation: str, labelnames: list[str] | None = None, registry=None) -> None:
        self.name = name
        self.documentation = documentation
        self.labelnames = tuple(labelnames or [])
        self.registry = registry
        self.samples: dict[tuple[tuple[str, str], ...], _MetricChild] = {}
        self.root = _MetricChild()

    def labels(self, **labels: str) -> _MetricChild:
        key = tuple(sorted(labels.items()))
        return self.samples.setdefault(key, _MetricChild())

    def inc(self, amount: float = 1.0) -> None:
        self.root.inc(amount)

    def dec(self, amount: float = 1.0) -> None:
        self.root.dec(amount)

    def observe(self, amount: float) -> None:
        self.root.observe(amount)


class _Counter(_MetricBase):
    pass


class _Histogram(_MetricBase):
    pass


class _Gauge(_MetricBase):
    pass


class _CollectorRegistry:
    def __init__(self, *, auto_describe: bool) -> None:
        self.auto_describe = auto_describe


def _fake_prometheus_client_module() -> types.ModuleType:
    module = types.ModuleType("prometheus_client")
    setattr(module, "Counter", _Counter)
    setattr(module, "Histogram", _Histogram)
    setattr(module, "Gauge", _Gauge)
    setattr(module, "CollectorRegistry", _CollectorRegistry)
    return module


@pytest.fixture(autouse=True)
def reset_metrics_modules() -> Iterator[None]:
    module_names = ("app.core.metrics.prometheus", "app.core.metrics.middleware")
    previous = {name: sys.modules.get(name) for name in module_names}
    try:
        yield
    finally:
        for name in module_names:
            sys.modules.pop(name, None)
        for name, module in previous.items():
            if module is not None:
                sys.modules[name] = module


def _load_metrics_modules(
    monkeypatch: pytest.MonkeyPatch, *, prometheus_client_module: types.ModuleType | None
) -> tuple[types.ModuleType, types.ModuleType]:
    for name in ("app.core.metrics.prometheus", "app.core.metrics.middleware"):
        sys.modules.pop(name, None)

    if prometheus_client_module is not None:
        monkeypatch.setitem(sys.modules, "prometheus_client", prometheus_client_module)
    else:
        monkeypatch.delitem(sys.modules, "prometheus_client", raising=False)
        real_import = builtins.__import__

        def _missing_prometheus_import(name, globals=None, locals=None, fromlist=(), level=0):
            if name == "prometheus_client":
                raise ImportError("prometheus_client is not installed")
            return real_import(name, globals, locals, fromlist, level)

        monkeypatch.setattr(builtins, "__import__", _missing_prometheus_import)

    prometheus_module = importlib.import_module("app.core.metrics.prometheus")
    middleware_module = importlib.import_module("app.core.metrics.middleware")
    return prometheus_module, middleware_module


def test_prometheus_metrics_defined_when_dependency_available(monkeypatch: pytest.MonkeyPatch) -> None:
    prometheus_module, _ = _load_metrics_modules(monkeypatch, prometheus_client_module=_fake_prometheus_client_module())

    assert prometheus_module.PROMETHEUS_AVAILABLE is True
    assert prometheus_module.REGISTRY is not None
    assert prometheus_module.requests_total.name == "codex_lb_requests_total"
    assert prometheus_module.request_duration_seconds.name == "codex_lb_request_duration_seconds"
    assert prometheus_module.active_connections.name == "codex_lb_active_connections"
    assert prometheus_module.bridge_instance_mismatch_total.name == "codex_lb_bridge_instance_mismatch_total"
    assert prometheus_module.bridge_instance_mismatch_total.labelnames == ("outcome",)


@pytest.mark.asyncio
async def test_metrics_middleware_records_request_metrics(monkeypatch: pytest.MonkeyPatch) -> None:
    prometheus_module, middleware_module = _load_metrics_modules(
        monkeypatch,
        prometheus_client_module=_fake_prometheus_client_module(),
    )

    app = FastAPI()
    app.add_middleware(middleware_module.MetricsMiddleware, enabled=True)

    @app.get("/v1/chat/completions/123")
    async def tracked_route() -> dict[str, str]:
        return {"status": "ok"}

    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://testserver") as client:
        response = await client.get("/v1/chat/completions/123")

    assert response.status_code == 200
    request_sample = prometheus_module.requests_total.samples[
        (("method", "GET"), ("path", "/v1/..."), ("status", "200"))
    ]
    duration_sample = prometheus_module.request_duration_seconds.samples[(("method", "GET"), ("path", "/v1/..."))]
    assert request_sample.value == 1.0
    assert len(duration_sample.observations) == 1
    assert prometheus_module.active_connections.root.value == 0.0


@pytest.mark.asyncio
async def test_metrics_middleware_noops_without_prometheus_client(monkeypatch: pytest.MonkeyPatch) -> None:
    prometheus_module, middleware_module = _load_metrics_modules(monkeypatch, prometheus_client_module=None)

    app = FastAPI()
    app.add_middleware(middleware_module.MetricsMiddleware, enabled=True)

    @app.get("/health")
    async def health() -> dict[str, str]:
        return {"status": "ok"}

    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://testserver") as client:
        response = await client.get("/health")

    assert response.status_code == 200
    assert response.json() == {"status": "ok"}
    assert prometheus_module.PROMETHEUS_AVAILABLE is False
    assert prometheus_module.requests_total is None


def test_bridge_instance_mismatch_counter_increments(monkeypatch: pytest.MonkeyPatch) -> None:
    prometheus_module, _ = _load_metrics_modules(monkeypatch, prometheus_client_module=_fake_prometheus_client_module())

    assert prometheus_module.PROMETHEUS_AVAILABLE is True
    counter = prometheus_module.bridge_instance_mismatch_total
    assert counter is not None

    fallback_sample = counter.labels(outcome="fallback")
    assert fallback_sample.value == 0.0

    fallback_sample.inc()
    assert fallback_sample.value == 1.0

    fallback_sample.inc()
    assert fallback_sample.value == 2.0


def test_bridge_instance_mismatch_counter_noop_without_prometheus(monkeypatch: pytest.MonkeyPatch) -> None:
    prometheus_module, _ = _load_metrics_modules(monkeypatch, prometheus_client_module=None)

    assert prometheus_module.PROMETHEUS_AVAILABLE is False
    assert prometheus_module.bridge_instance_mismatch_total is None



from __future__ import annotations

import pytest

import app.modules.proxy.service as proxy_module


@pytest.mark.asyncio
@pytest.mark.e2e
async def test_proxy_chat_completions_flow(
    client,
    monkeypatch,
    setup_dashboard_password,
    enable_api_key_auth,
    create_api_key,
    import_test_account,
):
    await setup_dashboard_password(client)
    await enable_api_key_auth(client)
    created = await create_api_key(
        client,
        name="e2e-proxy-key",
        limits=[
            {"limitType": "total_tokens", "limitWindow": "weekly", "maxValue": 1000},
        ],
    )
    await import_test_account(
        client,
        account_id="acc_e2e_proxy",
        email="e2e-proxy@example.com",
    )

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        yield 'data: {"type":"response.output_text.delta","delta":"hi"}\n\n'
        yield (
            'data: {"type":"response.completed","response":{"id":"resp_e2e_proxy","usage":'
            '{"input_tokens":2,"output_tokens":3,"total_tokens":5}}}\n\n'
        )

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    response = await client.post(
        "/v1/chat/completions",
        headers={"Authorization": f"Bearer {created['key']}"},
        json={
            "model": "gpt-5.2",
            "messages": [{"role": "user", "content": "hi"}],
        },
    )
    assert response.status_code == 200
    payload = response.json()
    assert payload["id"] == "resp_e2e_proxy"
    assert payload["object"] == "chat.completion"
    assert payload["usage"]["total_tokens"] == 5

    listed = await client.get("/api/api-keys/")
    assert listed.status_code == 200
    row = next(item for item in listed.json() if item["id"] == created["id"])
    assert row["usageSummary"]["requestCount"] == 1
    assert row["usageSummary"]["totalTokens"] == 5



from __future__ import annotations

import base64
import json
from datetime import timedelta, timezone
from types import SimpleNamespace

import pytest
from sqlalchemy import text

import app.modules.proxy.provider_adapters as provider_adapters_module
import app.modules.proxy.service as proxy_module
from app.core.crypto import TokenEncryptor
from app.core.openai.models import OpenAIResponsePayload
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus
from app.db.session import SessionLocal
from app.modules.accounts.repository import AccountsRepository
from app.modules.usage.repository import UsageRepository

pytestmark = pytest.mark.integration


class _SettingsCache:
    def __init__(self, settings: object) -> None:
        self._settings = settings

    async def get(self) -> object:
        return self._settings


def _encode_jwt(payload: dict) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


def _make_auth_json(account_id: str, email: str) -> dict:
    payload = {
        "email": email,
        "chatgpt_account_id": account_id,
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    return {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access-token",
            "refreshToken": "refresh-token",
            "accountId": account_id,
        },
    }


async def _import_account(async_client, account_id: str, email: str) -> str:
    auth_json = _make_auth_json(account_id, email)
    files = {"auth_json": ("auth.json", json.dumps(auth_json), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200
    payload = response.json()
    return payload["accountId"]


async def _set_routing_settings(
    async_client,
    monkeypatch: pytest.MonkeyPatch,
    *,
    sticky_threads_enabled: bool,
    prefer_earlier_reset_accounts: bool = False,
    routing_strategy: str = "usage_weighted",
) -> None:
    response = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": sticky_threads_enabled,
            "preferEarlierResetAccounts": prefer_earlier_reset_accounts,
            "routingStrategy": routing_strategy,
        },
    )
    assert response.status_code == 200
    _install_proxy_settings_cache(
        monkeypatch,
        sticky_threads_enabled=sticky_threads_enabled,
        prefer_earlier_reset_accounts=prefer_earlier_reset_accounts,
        routing_strategy=routing_strategy,
    )


def _install_proxy_settings_cache(
    monkeypatch: pytest.MonkeyPatch,
    *,
    sticky_threads_enabled: bool,
    prefer_earlier_reset_accounts: bool = False,
    routing_strategy: str = "usage_weighted",
    openai_cache_affinity_max_age_seconds: int = 300,
    sticky_reallocation_budget_threshold_pct: float = 95.0,
    openai_prompt_cache_key_derivation_enabled: bool = True,
) -> None:
    settings = SimpleNamespace(
        prefer_earlier_reset_accounts=prefer_earlier_reset_accounts,
        sticky_threads_enabled=sticky_threads_enabled,
        openai_cache_affinity_max_age_seconds=openai_cache_affinity_max_age_seconds,
        sticky_reallocation_budget_threshold_pct=sticky_reallocation_budget_threshold_pct,
        openai_prompt_cache_key_derivation_enabled=openai_prompt_cache_key_derivation_enabled,
        routing_strategy=routing_strategy,
        proxy_request_budget_seconds=75.0,
        compact_request_budget_seconds=75.0,
        transcription_request_budget_seconds=120.0,
        upstream_compact_timeout_seconds=None,
        upstream_stream_transport="auto",
        log_proxy_request_payload=False,
        log_proxy_request_shape=False,
        log_proxy_request_shape_raw_cache_key=False,
        log_proxy_service_tier_trace=False,
        http_responses_session_bridge_enabled=False,
        http_responses_session_bridge_idle_ttl_seconds=120.0,
        http_responses_session_bridge_max_sessions=128,
        http_responses_session_bridge_queue_limit=8,
    )
    monkeypatch.setattr(proxy_module, "get_settings_cache", lambda: _SettingsCache(settings))
    monkeypatch.setattr(proxy_module, "get_settings", lambda: settings)


@pytest.mark.asyncio
async def test_proxy_stream_sticky_threads_reallocate_by_prompt_cache_key(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=True)
    acc_a_id = await _import_account(async_client, "acc_a", "a@example.com")
    acc_b_id = await _import_account(async_client, "acc_b", "b@example.com")

    seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_1"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    payload = {
        "model": "gpt-5.1",
        "instructions": "hi",
        "input": [],
        "stream": True,
        "prompt_cache_key": "thread_123",
    }

    response = await async_client.post("/backend-api/codex/responses", json=payload)
    assert response.status_code == 200

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=95.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    response = await async_client.post("/backend-api/codex/responses", json=payload)
    assert response.status_code == 200

    assert seen == ["acc_a", "acc_a"]


@pytest.mark.asyncio
async def test_proxy_sticky_switches_when_pinned_rate_limited(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=True)
    encryptor = TokenEncryptor()
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    acc_a = Account(
        id="acc_rl_a",
        chatgpt_account_id="acc_rl_a",
        email="rl_a@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access-a"),
        refresh_token_encrypted=encryptor.encrypt("refresh-a"),
        id_token_encrypted=encryptor.encrypt("id-a"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )
    acc_b = Account(
        id="acc_rl_b",
        chatgpt_account_id="acc_rl_b",
        email="rl_b@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access-b"),
        refresh_token_encrypted=encryptor.encrypt("refresh-b"),
        id_token_encrypted=encryptor.encrypt("id-b"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)
        await accounts_repo.upsert(acc_a)
        await accounts_repo.upsert(acc_b)
        await usage_repo.add_entry(
            account_id=acc_a.id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b.id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        seen.append(account_id)
        if account_id == acc_a.id:
            yield (
                'data: {"type":"response.failed","response":{"error":{"code":"rate_limit_exceeded",'
                '"message":"slow down"}}}\n\n'
            )
            return
        yield 'data: {"type":"response.completed","response":{"id":"resp_ok"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.1",
        "instructions": "hi",
        "input": [],
        "stream": True,
        "prompt_cache_key": "thread_rl",
    }
    response = await async_client.post("/backend-api/codex/responses", json=payload)
    assert response.status_code == 200

    # First attempt is pinned acc_a, which rate limits; retry should switch to acc_b and update stickiness.
    assert seen[:2] == [acc_a.id, acc_b.id]


@pytest.mark.asyncio
async def test_proxy_compact_reallocates_sticky_mapping(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=True)
    acc_c1_id = await _import_account(async_client, "acc_c1", "c1@example.com")
    acc_c2_id = await _import_account(async_client, "acc_c2", "c2@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_c1_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_c2_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    stream_seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        stream_seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_1"}}\n\n'

    compact_seen: list[str] = []

    async def fake_compact(payload, headers, access_token, account_id):
        compact_seen.append(account_id)
        return OpenAIResponsePayload.model_validate({"output": []})

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)
    monkeypatch.setattr(provider_adapters_module, "core_compact_responses", fake_compact)

    thread_key = "thread_compact_1"
    stream_payload = {
        "model": "gpt-5.1",
        "instructions": "hi",
        "input": [],
        "stream": True,
        "prompt_cache_key": thread_key,
    }
    response = await async_client.post("/backend-api/codex/responses", json=stream_payload)
    assert response.status_code == 200
    assert stream_seen == ["acc_c1"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_c1_id,
            used_percent=90.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_c2_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    compact_payload = {
        "model": "gpt-5.1",
        "instructions": "summarize",
        "input": [],
        "prompt_cache_key": thread_key,  # extra field accepted by ResponsesCompactRequest (extra="allow")
    }
    response = await async_client.post("/backend-api/codex/responses/compact", json=compact_payload)
    assert response.status_code == 200
    assert compact_seen == ["acc_c1"]

    response = await async_client.post("/backend-api/codex/responses", json=stream_payload)
    assert response.status_code == 200
    assert stream_seen == ["acc_c1", "acc_c1"]


@pytest.mark.asyncio
async def test_proxy_codex_session_id_pins_responses_and_compact_without_sticky_threads(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=False)
    acc_a_id = await _import_account(async_client, "acc_sid_a", "sid_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_sid_b", "sid_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    stream_seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        stream_seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_session"}}\n\n'

    compact_seen: list[str] = []

    async def fake_compact(payload, headers, access_token, account_id):
        compact_seen.append(account_id)
        return OpenAIResponsePayload.model_validate({"output": []})

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)
    monkeypatch.setattr(provider_adapters_module, "core_compact_responses", fake_compact)

    headers = {"session_id": "codex-thread-123"}
    stream_payload = {
        "model": "gpt-5.1",
        "instructions": "hi",
        "input": [],
        "stream": True,
    }
    response = await async_client.post("/backend-api/codex/responses", json=stream_payload, headers=headers)
    assert response.status_code == 200
    assert stream_seen == ["acc_sid_a"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=95.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    compact_payload = {
        "model": "gpt-5.1",
        "instructions": "summarize",
        "input": [{"role": "user", "content": [{"type": "input_text", "text": "hello"}]}],
    }
    response = await async_client.post(
        "/backend-api/codex/responses/compact",
        json=compact_payload,
        headers=headers,
    )
    assert response.status_code == 200
    assert compact_seen == ["acc_sid_a"]

    response = await async_client.post("/backend-api/codex/responses", json=stream_payload, headers=headers)
    assert response.status_code == 200
    assert stream_seen == ["acc_sid_a", "acc_sid_a"]


@pytest.mark.asyncio
async def test_proxy_codex_session_id_compact_first_pins_followup_stream_without_sticky_threads(
    async_client,
    monkeypatch,
):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=False)
    acc_a_id = await _import_account(async_client, "acc_sid_compact_a", "sid_compact_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_sid_compact_b", "sid_compact_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    compact_seen: list[str] = []

    async def fake_compact(payload, headers, access_token, account_id):
        compact_seen.append(account_id)
        return OpenAIResponsePayload.model_validate({"output": []})

    stream_seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        stream_seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_compact_first"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_compact_responses", fake_compact)
    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    headers = {"session_id": "codex-compact-first-123"}
    compact_payload = {
        "model": "gpt-5.1",
        "instructions": "summarize",
        "input": [{"role": "user", "content": [{"type": "input_text", "text": "hello"}]}],
    }
    response = await async_client.post(
        "/backend-api/codex/responses/compact",
        json=compact_payload,
        headers=headers,
    )
    assert response.status_code == 200
    assert compact_seen == ["acc_sid_compact_a"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=95.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    stream_payload = {
        "model": "gpt-5.1",
        "instructions": "continue",
        "input": [],
        "stream": True,
    }
    response = await async_client.post("/backend-api/codex/responses", json=stream_payload, headers=headers)
    assert response.status_code == 200
    assert stream_seen == ["acc_sid_compact_a"]


@pytest.mark.asyncio
async def test_proxy_codex_session_id_switches_when_pinned_rate_limited(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=False)
    acc_a_id = await _import_account(async_client, "acc_sid_retry_a", "sid_retry_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_sid_retry_b", "sid_retry_b@example.com")
    upstream_acc_a = "acc_sid_retry_a"
    upstream_acc_b = "acc_sid_retry_b"

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    stream_seen: list[str] = []
    fail_pinned = {"enabled": False}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        stream_seen.append(account_id)
        if account_id == upstream_acc_a and fail_pinned["enabled"]:
            yield (
                'data: {"type":"response.failed","response":{"error":{"code":"rate_limit_exceeded",'
                '"message":"slow down"}}}\n\n'
            )
            return
        yield 'data: {"type":"response.completed","response":{"id":"resp_session_retry"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    headers = {"session_id": "codex-session-retry-123"}
    stream_payload = {
        "model": "gpt-5.1",
        "instructions": "hi",
        "input": [],
        "stream": True,
    }
    response = await async_client.post("/backend-api/codex/responses", json=stream_payload, headers=headers)
    assert response.status_code == 200
    assert stream_seen == [upstream_acc_a]

    fail_pinned["enabled"] = True
    response = await async_client.post("/backend-api/codex/responses", json=stream_payload, headers=headers)
    assert response.status_code == 200

    response = await async_client.post("/backend-api/codex/responses", json=stream_payload, headers=headers)
    assert response.status_code == 200
    assert stream_seen == [upstream_acc_a, upstream_acc_a, upstream_acc_b, upstream_acc_b]


@pytest.mark.asyncio
async def test_v1_session_id_does_not_create_durable_codex_session_affinity(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=False)
    acc_a_id = await _import_account(async_client, "acc_v1_sid_a", "v1_sid_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_v1_sid_b", "v1_sid_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    stream_seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        stream_seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_v1_session"}}\n\n'

    compact_seen: list[str] = []

    async def fake_compact(payload, headers, access_token, account_id):
        compact_seen.append(account_id)
        return OpenAIResponsePayload.model_validate({"output": []})

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)
    monkeypatch.setattr(provider_adapters_module, "core_compact_responses", fake_compact)

    headers = {"session_id": "v1-thread-123"}
    stream_payload = {
        "model": "gpt-5.1",
        "input": "hello",
        "stream": True,
    }
    response = await async_client.post("/v1/responses", json=stream_payload, headers=headers)
    assert response.status_code == 200
    assert stream_seen == ["acc_v1_sid_a"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=95.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    compact_payload = {
        "model": "gpt-5.1",
        "input": "hello",
    }
    response = await async_client.post("/v1/responses/compact", json=compact_payload, headers=headers)
    assert response.status_code == 200
    assert compact_seen == ["acc_v1_sid_a"]

    async with SessionLocal() as session:
        codex_row = (
            await session.execute(
                text("SELECT kind FROM sticky_sessions WHERE key = :key"),
                {"key": "v1-thread-123"},
            )
        ).fetchone()
        assert codex_row is None


@pytest.mark.asyncio
async def test_v1_prompt_cache_key_reuses_recent_responses_and_compact_without_sticky_threads(
    async_client,
    monkeypatch,
):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=False)
    _install_proxy_settings_cache(
        monkeypatch,
        sticky_threads_enabled=False,
        openai_cache_affinity_max_age_seconds=60,
    )
    acc_a_id = await _import_account(async_client, "acc_v1_cache_a", "v1_cache_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_v1_cache_b", "v1_cache_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    stream_seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        stream_seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_v1_cache"}}\n\n'

    compact_seen: list[str] = []

    async def fake_compact(payload, headers, access_token, account_id):
        compact_seen.append(account_id)
        return OpenAIResponsePayload.model_validate({"output": []})

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)
    monkeypatch.setattr(provider_adapters_module, "core_compact_responses", fake_compact)

    thread_key = "v1-cache-thread-123"
    stream_payload = {
        "model": "gpt-5.1",
        "input": "hello",
        "stream": True,
        "prompt_cache_key": thread_key,
    }
    response = await async_client.post("/v1/responses", json=stream_payload)
    assert response.status_code == 200
    assert stream_seen == ["acc_v1_cache_a"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=95.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    compact_payload = {
        "model": "gpt-5.1",
        "input": "hello",
        "prompt_cache_key": thread_key,
    }
    response = await async_client.post("/v1/responses/compact", json=compact_payload)
    assert response.status_code == 200
    assert compact_seen == ["acc_v1_cache_a"]

    response = await async_client.post("/v1/responses", json=stream_payload)
    assert response.status_code == 200
    assert stream_seen == ["acc_v1_cache_a", "acc_v1_cache_a"]


@pytest.mark.asyncio
async def test_v1_responses_derives_prompt_cache_key_when_absent(async_client, monkeypatch):
    _install_proxy_settings_cache(
        monkeypatch,
        sticky_threads_enabled=False,
        openai_cache_affinity_max_age_seconds=60,
    )
    acc_a_id = await _import_account(async_client, "acc_v1_derived_a", "v1_derived_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_v1_derived_b", "v1_derived_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen_keys: list[str | None] = []
    seen_accounts: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        seen_accounts.append(account_id)
        seen_keys.append(getattr(payload, "prompt_cache_key", None))
        yield 'data: {"type":"response.completed","response":{"id":"resp_v1_derived"}}\\n\\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.1", "input": "hello", "stream": True}
    response = await async_client.post("/v1/responses", json=payload)
    assert response.status_code == 200
    assert seen_accounts == ["acc_v1_derived_a"]
    assert isinstance(seen_keys[0], str)
    assert seen_keys[0]


@pytest.mark.asyncio
async def test_backend_codex_session_affinity_also_forwards_prompt_cache_key_when_missing(async_client, monkeypatch):
    _install_proxy_settings_cache(monkeypatch, sticky_threads_enabled=False)
    acc_id = await _import_account(async_client, "acc_codex_sid_1", "codex_sid_1@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen_keys: list[str | None] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        seen_keys.append(getattr(payload, "prompt_cache_key", None))
        yield 'data: {"type":"response.completed","response":{"id":"resp_backend_codex"}}\\n\\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    response = await async_client.post(
        "/backend-api/codex/responses",
        json={
            "model": "gpt-5.1",
            "instructions": "hi",
            "input": [{"role": "user", "content": [{"type": "input_text", "text": "hello"}]}],
            "stream": True,
        },
        headers={"session_id": "backend-thread-123"},
    )
    assert response.status_code == 200
    assert isinstance(seen_keys[0], str)
    assert seen_keys[0]

    async with SessionLocal() as session:
        row = (
            await session.execute(
                text("SELECT kind FROM sticky_sessions WHERE key = :key"),
                {"key": "backend-thread-123"},
            )
        ).fetchone()
        assert row is not None
        assert row[0] == "codex_session"


@pytest.mark.asyncio
async def test_backend_responses_http_forwards_previous_response_id(async_client, monkeypatch):
    _install_proxy_settings_cache(monkeypatch, sticky_threads_enabled=False)
    acc_id = await _import_account(async_client, "acc_prev_http_1", "prev_http_1@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen_prev_ids: list[str | None] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del headers, access_token, account_id, base_url, raise_for_status, _kw
        seen_prev_ids.append(getattr(payload, "previous_response_id", None))
        yield 'data: {"type":"response.completed","response":{"id":"resp_prev_http"}}\\n\\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    response = await async_client.post(
        "/backend-api/codex/responses",
        json={
            "model": "gpt-5.1",
            "instructions": "hi",
            "previous_response_id": "resp_prev_http_123",
            "input": [{"role": "user", "content": [{"type": "input_text", "text": "continue"}]}],
            "stream": True,
        },
        headers={"session_id": "backend-thread-prev-http-123"},
    )
    assert response.status_code == 200
    assert seen_prev_ids == ["resp_prev_http_123"]


@pytest.mark.asyncio
async def test_v1_responses_http_forwards_previous_response_id(async_client, monkeypatch):
    _install_proxy_settings_cache(monkeypatch, sticky_threads_enabled=False)
    acc_id = await _import_account(async_client, "acc_v1_prev_http_1", "v1_prev_http_1@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen_prev_ids: list[str | None] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del headers, access_token, account_id, base_url, raise_for_status, _kw
        seen_prev_ids.append(getattr(payload, "previous_response_id", None))
        yield 'data: {"type":"response.completed","response":{"id":"resp_v1_prev_http"}}\\n\\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    response = await async_client.post(
        "/v1/responses",
        json={
            "model": "gpt-5.1",
            "input": "continue",
            "previous_response_id": "resp_prev_v1_http_123",
            "stream": True,
        },
    )
    assert response.status_code == 200
    assert seen_prev_ids == ["resp_prev_v1_http_123"]


@pytest.mark.asyncio
async def test_v1_prompt_cache_key_rebalances_after_affinity_expires(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=False)
    _install_proxy_settings_cache(
        monkeypatch,
        sticky_threads_enabled=False,
        openai_cache_affinity_max_age_seconds=60,
    )
    acc_a_id = await _import_account(async_client, "acc_v1_expire_a", "v1_expire_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_v1_expire_b", "v1_expire_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    stream_seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        stream_seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_v1_cache_expire"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    thread_key = "v1-cache-expire-thread-123"
    stream_payload = {
        "model": "gpt-5.1",
        "input": "hello",
        "stream": True,
        "prompt_cache_key": thread_key,
    }
    response = await async_client.post("/v1/responses", json=stream_payload)
    assert response.status_code == 200
    assert stream_seen == ["acc_v1_expire_a"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=95.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        stale_updated_at = utcnow() - timedelta(minutes=10)
        await session.execute(
            text(
                """
                UPDATE sticky_sessions
                SET updated_at = :stale_updated_at
                WHERE key = :sticky_key AND kind = 'prompt_cache'
                """
            ),
            {"sticky_key": thread_key, "stale_updated_at": stale_updated_at},
        )
        await session.commit()

    response = await async_client.post("/v1/responses", json=stream_payload)
    assert response.status_code == 200
    assert stream_seen == ["acc_v1_expire_a", "acc_v1_expire_b"]


@pytest.mark.asyncio
async def test_codex_endpoint_uses_prompt_cache_sticky_kind(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=True)
    acc_id = await _import_account(async_client, "acc_kind_a", "kind_a@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_k"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True, "prompt_cache_key": "pck_abc"}
    await async_client.post("/backend-api/codex/responses", json=payload)
    assert seen == ["acc_kind_a"]

    async with SessionLocal() as session:
        row = (await session.execute(text("SELECT kind FROM sticky_sessions WHERE key = 'pck_abc'"))).fetchone()
        assert row is not None
        assert row[0] == "prompt_cache"


@pytest.mark.asyncio
async def test_v1_auto_derived_key_separates_parallel_sessions(async_client, monkeypatch):
    _install_proxy_settings_cache(monkeypatch, sticky_threads_enabled=False)
    acc_a_id = await _import_account(async_client, "acc_par_a", "par_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_par_b", "par_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_p"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    session_a = {"model": "gpt-5.1", "input": "build a server", "stream": True}
    session_b = {"model": "gpt-5.1", "input": "write tests", "stream": True}

    await async_client.post("/v1/responses", json=session_a)
    await async_client.post("/v1/responses", json=session_b)

    assert len(seen) == 2
    assert seen[0] == "acc_par_a"


@pytest.mark.asyncio
async def test_v1_auto_derived_key_stable_across_turns(async_client, monkeypatch):
    _install_proxy_settings_cache(monkeypatch, sticky_threads_enabled=False)
    acc_a_id = await _import_account(async_client, "acc_turn_a", "turn_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_turn_b", "turn_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_t"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    turn1 = {
        "model": "gpt-5.1",
        "input": [{"role": "user", "content": "build a server"}],
        "stream": True,
    }
    turn2 = {
        "model": "gpt-5.1",
        "input": [
            {"role": "user", "content": "build a server"},
            {"role": "assistant", "content": "Here is a server..."},
            {"role": "user", "content": "add logging"},
        ],
        "stream": True,
    }

    await async_client.post("/v1/responses", json=turn1)
    assert seen == ["acc_turn_a"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=90.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    await async_client.post("/v1/responses", json=turn2)

    assert seen == ["acc_turn_a", "acc_turn_a"]


@pytest.mark.asyncio
async def test_reallocate_sticky_respects_existing_session_then_falls_back(async_client, monkeypatch):
    await _set_routing_settings(async_client, monkeypatch, sticky_threads_enabled=True)
    acc_a_id = await _import_account(async_client, "acc_realloc_a", "realloc_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_realloc_b", "realloc_b@example.com")

    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_r"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True, "prompt_cache_key": "realloc_key"}
    await async_client.post("/backend-api/codex/responses", json=payload)
    assert seen == ["acc_realloc_a"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=95.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    await async_client.post("/backend-api/codex/responses", json=payload)
    assert seen == ["acc_realloc_a", "acc_realloc_a"]

    async with SessionLocal() as session:
        await session.execute(text("DELETE FROM accounts WHERE chatgpt_account_id = 'acc_realloc_a'"))
        await session.commit()

    await async_client.post("/backend-api/codex/responses", json=payload)
    assert len(seen) == 3
    assert seen[2] == "acc_realloc_b"



from __future__ import annotations

from app.modules.shared.schemas import DashboardModel


class DashboardAuthSessionResponse(DashboardModel):
    authenticated: bool
    password_required: bool
    totp_required_on_login: bool
    totp_configured: bool
    bootstrap_required: bool = False
    bootstrap_token_configured: bool = False


class TotpSetupStartResponse(DashboardModel):
    secret: str
    otpauth_uri: str
    qr_svg_data_uri: str


class TotpSetupConfirmRequest(DashboardModel):
    secret: str
    code: str


class TotpVerifyRequest(DashboardModel):
    code: str


class PasswordSetupRequest(DashboardModel):
    password: str
    bootstrap_token: str | None = None


class PasswordLoginRequest(DashboardModel):
    password: str


class PasswordChangeRequest(DashboardModel):
    current_password: str
    new_password: str


class PasswordRemoveRequest(DashboardModel):
    password: str



"""add credit-based api key limit enum values

Revision ID: 20260403_000000_add_credit_api_key_limit_values
Revises: 20260402_000000_switch_dashboard_routing_default_to_capacity_weighted
Create Date: 2026-04-03
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

revision = "20260403_000000_add_credit_api_key_limit_values"
down_revision = "20260402_000000_switch_dashboard_routing_default_to_capacity_weighted"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def upgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "api_key_limits"):
        return
    if bind.dialect.name != "postgresql":
        return

    op.execute(sa.text("ALTER TYPE limit_type ADD VALUE IF NOT EXISTS 'credits'"))
    op.execute(sa.text("ALTER TYPE limit_window ADD VALUE IF NOT EXISTS '5h'"))
    op.execute(sa.text("ALTER TYPE limit_window ADD VALUE IF NOT EXISTS '7d'"))


def downgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "api_key_limits"):
        return
    if bind.dialect.name != "postgresql":
        return
    # PostgreSQL enum values cannot be removed safely in-place here.
    return



from __future__ import annotations

from fastapi import APIRouter, Body, Depends, Request
from fastapi.responses import JSONResponse

from app.core.auth.dependencies import insecure_remote_no_auth_allowed, set_dashboard_error_format
from app.core.config.settings import get_settings
from app.core.config.settings_cache import get_settings_cache
from app.core.exceptions import (
    DashboardAuthError,
    DashboardBadRequestError,
    DashboardConflictError,
    DashboardRateLimitError,
    DashboardValidationError,
)
from app.core.request_locality import is_local_request
from app.dependencies import DashboardAuthContext, get_dashboard_auth_context
from app.modules.dashboard_auth.schemas import (
    DashboardAuthSessionResponse,
    PasswordChangeRequest,
    PasswordLoginRequest,
    PasswordRemoveRequest,
    PasswordSetupRequest,
    TotpSetupConfirmRequest,
    TotpSetupStartResponse,
    TotpVerifyRequest,
)
from app.modules.dashboard_auth.service import (
    DASHBOARD_SESSION_COOKIE,
    InvalidCredentialsError,
    PasswordAlreadyConfiguredError,
    PasswordNotConfiguredError,
    PasswordSessionRequiredError,
    TotpAlreadyConfiguredError,
    TotpInvalidCodeError,
    TotpInvalidSetupError,
    TotpNotConfiguredError,
    get_dashboard_session_store,
    get_password_rate_limiter,
    get_totp_rate_limiter,
)

router = APIRouter(
    prefix="/api/dashboard-auth",
    tags=["dashboard"],
    dependencies=[Depends(set_dashboard_error_format)],
)


def _session_client_key(request: Request, *, prefix: str) -> str:
    return f"{prefix}:{request.client.host if request.client else 'unknown'}"


async def _has_active_password_session(request: Request, context: DashboardAuthContext) -> bool:
    settings = await context.repository.get_settings()
    if settings.password_hash is None:
        return False
    session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
    return get_dashboard_session_store().is_password_verified(session_id)


async def _validate_password_management_session(request: Request) -> None:
    session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
    session_state = get_dashboard_session_store().get(session_id)
    if session_state is None or not session_state.password_verified:
        raise DashboardAuthError("Authentication is required")

    settings = await get_settings_cache().get()
    if settings.totp_required_on_login and not session_state.totp_verified:
        raise DashboardAuthError(
            "TOTP verification is required for dashboard access",
            code="totp_required",
        )


@router.get("/session", response_model=DashboardAuthSessionResponse)
async def get_dashboard_auth_session(
    request: Request,
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> DashboardAuthSessionResponse:
    session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
    response = await context.service.get_session_state(session_id)
    if (
        insecure_remote_no_auth_allowed(request)
        and not response.password_required
        and not response.totp_required_on_login
    ):
        return response.model_copy(
            update={
                "authenticated": True,
                "bootstrap_required": False,
                "bootstrap_token_configured": False,
                "totp_required_on_login": False,
            }
        )
    if response.password_required or is_local_request(request):
        return response
    bootstrap_token_configured = bool((get_settings().dashboard_bootstrap_token or "").strip())
    return response.model_copy(
        update={
            "authenticated": False,
            "bootstrap_required": True,
            "bootstrap_token_configured": bootstrap_token_configured,
        }
    )


@router.post("/password/setup", response_model=DashboardAuthSessionResponse)
async def setup_password(
    request: Request,
    payload: PasswordSetupRequest = Body(...),
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> DashboardAuthSessionResponse | JSONResponse:
    current_settings = await context.repository.get_settings()
    if current_settings.password_hash is None and not is_local_request(request):
        configured_bootstrap_token = (get_settings().dashboard_bootstrap_token or "").strip()
        if not configured_bootstrap_token:
            raise DashboardAuthError(
                "Remote bootstrap is disabled until CODEX_LB_DASHBOARD_BOOTSTRAP_TOKEN is configured.",
                code="bootstrap_unavailable",
            )
        submitted_bootstrap_token = (payload.bootstrap_token or "").strip()
        if submitted_bootstrap_token != configured_bootstrap_token:
            raise DashboardAuthError("Invalid dashboard bootstrap token.", code="invalid_bootstrap_token")
    password = payload.password.strip()
    if len(password) < 8:
        raise DashboardValidationError("Password must be at least 8 characters")
    try:
        await context.service.setup_password(password)
    except PasswordAlreadyConfiguredError as exc:
        raise DashboardConflictError(str(exc), code="password_already_configured") from exc

    await get_settings_cache().invalidate()
    session_id = get_dashboard_session_store().create(password_verified=True, totp_verified=False)
    response = await context.service.get_session_state(session_id)
    json_response = JSONResponse(status_code=200, content=response.model_dump(by_alias=True))
    _set_session_cookie(json_response, session_id, request)
    return json_response


@router.post("/password/login", response_model=DashboardAuthSessionResponse)
async def login_password(
    request: Request,
    payload: PasswordLoginRequest = Body(...),
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> DashboardAuthSessionResponse | JSONResponse:
    settings = await get_settings_cache().get()
    if settings.password_hash is None:
        raise DashboardBadRequestError("Password is not configured", code="password_not_configured")

    limiter = get_password_rate_limiter()
    rate_key = _session_client_key(request, prefix="password_login")
    try:
        await limiter.check_and_increment(rate_key, context.session)
    except DashboardRateLimitError as exc:
        raise DashboardRateLimitError(
            f"Too many attempts. Try again in {exc.retry_after} seconds.",
            retry_after=exc.retry_after,
            code="password_rate_limited",
        ) from exc

    try:
        await context.service.verify_password(
            payload.password, actor_ip=request.client.host if request.client else None
        )
    except InvalidCredentialsError as exc:
        raise DashboardAuthError(str(exc), code="invalid_credentials") from exc
    except PasswordNotConfiguredError as exc:
        await limiter.clear_for_key(rate_key, context.session)
        raise DashboardBadRequestError(str(exc), code="password_not_configured") from exc

    await limiter.clear_for_key(rate_key, context.session)

    session_id = get_dashboard_session_store().create(password_verified=True, totp_verified=False)
    response = await context.service.get_session_state(session_id)
    json_response = JSONResponse(status_code=200, content=response.model_dump(by_alias=True))
    _set_session_cookie(json_response, session_id, request)
    return json_response


@router.post("/password/change")
async def change_password(
    request: Request,
    payload: PasswordChangeRequest = Body(...),
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> JSONResponse:
    await _validate_password_management_session(request)

    new_password = payload.new_password.strip()
    if len(new_password) < 8:
        raise DashboardValidationError("Password must be at least 8 characters")

    try:
        await context.service.change_password(payload.current_password, new_password)
    except PasswordNotConfiguredError as exc:
        raise DashboardBadRequestError(str(exc), code="password_not_configured") from exc
    except InvalidCredentialsError as exc:
        raise DashboardAuthError(str(exc), code="invalid_credentials") from exc

    await get_settings_cache().invalidate()
    return JSONResponse(status_code=200, content={"status": "ok"})


@router.delete("/password")
async def remove_password(
    request: Request,
    payload: PasswordRemoveRequest = Body(...),
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> JSONResponse:
    await _validate_password_management_session(request)

    try:
        await context.service.remove_password(payload.password)
    except PasswordNotConfiguredError as exc:
        raise DashboardBadRequestError(str(exc), code="password_not_configured") from exc
    except InvalidCredentialsError as exc:
        raise DashboardAuthError(str(exc), code="invalid_credentials") from exc

    await get_settings_cache().invalidate()
    response = JSONResponse(status_code=200, content={"status": "ok"})
    response.delete_cookie(key=DASHBOARD_SESSION_COOKIE, path="/")
    return response


@router.post("/totp/setup/start", response_model=TotpSetupStartResponse)
async def start_totp_setup(
    request: Request,
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> TotpSetupStartResponse:
    if not await _has_active_password_session(request, context):
        raise DashboardAuthError("Authentication is required")
    session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
    try:
        return await context.service.start_totp_setup(session_id=session_id)
    except PasswordSessionRequiredError as exc:
        raise DashboardAuthError(str(exc)) from exc
    except TotpAlreadyConfiguredError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_setup") from exc


@router.post("/totp/setup/confirm")
async def confirm_totp_setup(
    request: Request,
    payload: TotpSetupConfirmRequest = Body(...),
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> JSONResponse:
    if not await _has_active_password_session(request, context):
        raise DashboardAuthError("Authentication is required")

    limiter = get_totp_rate_limiter()
    rate_key = _session_client_key(request, prefix="totp_setup_confirm")
    try:
        await limiter.check_and_increment(rate_key, context.session)
    except DashboardRateLimitError as exc:
        raise DashboardRateLimitError(
            f"Too many attempts. Try again in {exc.retry_after} seconds.",
            retry_after=exc.retry_after,
            code="totp_rate_limited",
        ) from exc

    try:
        session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
        await context.service.confirm_totp_setup(
            session_id=session_id,
            secret=payload.secret,
            code=payload.code,
            actor_ip=request.client.host if request.client else None,
        )
    except PasswordSessionRequiredError as exc:
        raise DashboardAuthError(str(exc)) from exc
    except TotpInvalidCodeError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_code") from exc
    except TotpInvalidSetupError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_setup") from exc
    except TotpAlreadyConfiguredError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_setup") from exc

    await limiter.clear_for_key(rate_key, context.session)
    await get_settings_cache().invalidate()
    return JSONResponse(status_code=200, content={"status": "ok"})


@router.post("/totp/verify", response_model=DashboardAuthSessionResponse)
async def verify_totp(
    request: Request,
    payload: TotpVerifyRequest = Body(...),
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> DashboardAuthSessionResponse | JSONResponse:
    limiter = get_totp_rate_limiter()
    rate_key = _session_client_key(request, prefix="totp_verify")
    current_session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
    try:
        await context.service.ensure_active_password_session(current_session_id)
    except PasswordSessionRequiredError as exc:
        raise DashboardAuthError(str(exc)) from exc
    try:
        await limiter.check_and_increment(rate_key, context.session)
    except DashboardRateLimitError as exc:
        raise DashboardRateLimitError(
            f"Too many attempts. Try again in {exc.retry_after} seconds.",
            retry_after=exc.retry_after,
            code="totp_rate_limited",
        ) from exc
    try:
        session_id = await context.service.verify_totp(
            session_id=current_session_id,
            code=payload.code,
            actor_ip=request.client.host if request.client else None,
        )
    except PasswordSessionRequiredError as exc:
        raise DashboardAuthError(str(exc)) from exc
    except TotpInvalidCodeError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_code") from exc
    except TotpNotConfiguredError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_code") from exc

    await limiter.clear_for_key(rate_key, context.session)
    response = await context.service.get_session_state(session_id)
    json_response = JSONResponse(status_code=200, content=response.model_dump(by_alias=True))
    _set_session_cookie(json_response, session_id, request)
    return json_response


@router.post("/totp/disable")
async def disable_totp(
    request: Request,
    payload: TotpVerifyRequest = Body(...),
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> JSONResponse:
    limiter = get_totp_rate_limiter()
    rate_key = _session_client_key(request, prefix="totp_disable")
    session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
    try:
        await context.service.ensure_totp_verified_session(session_id)
    except PasswordSessionRequiredError as exc:
        raise DashboardAuthError(str(exc)) from exc
    try:
        await limiter.check_and_increment(rate_key, context.session)
    except DashboardRateLimitError as exc:
        raise DashboardRateLimitError(
            f"Too many attempts. Try again in {exc.retry_after} seconds.",
            retry_after=exc.retry_after,
            code="totp_rate_limited",
        ) from exc
    try:
        await context.service.disable_totp(
            session_id=session_id,
            code=payload.code,
            actor_ip=request.client.host if request.client else None,
        )
    except PasswordSessionRequiredError as exc:
        raise DashboardAuthError(str(exc)) from exc
    except TotpInvalidCodeError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_code") from exc
    except TotpNotConfiguredError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_code") from exc

    await limiter.clear_for_key(rate_key, context.session)
    await get_settings_cache().invalidate()
    return JSONResponse(status_code=200, content={"status": "ok"})


@router.post("/logout")
async def logout_dashboard(
    request: Request,
    context: DashboardAuthContext = Depends(get_dashboard_auth_context),
) -> JSONResponse:
    session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
    context.service.logout(session_id)
    response = JSONResponse(status_code=200, content={"status": "ok"})
    response.delete_cookie(key=DASHBOARD_SESSION_COOKIE, path="/")
    return response


def _set_session_cookie(response: JSONResponse, session_id: str, request: Request) -> None:
    response.set_cookie(
        key=DASHBOARD_SESSION_COOKIE,
        value=session_id,
        httponly=True,
        secure=request.url.scheme == "https",
        samesite="lax",
        max_age=12 * 60 * 60,
        path="/",
    )



from __future__ import annotations

from datetime import timedelta

import pytest

from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus, ApiKeyLimit, LimitType, LimitWindow, UsageHistory
from app.db.session import SessionLocal
from app.modules.api_keys.repository import ApiKeysRepository
from app.modules.api_keys.service import ApiKeyCreateData, ApiKeysService, LimitRuleInput
from app.modules.request_logs.repository import RequestLogsRepository

pytestmark = pytest.mark.integration


async def _create_api_key(
    *,
    name: str,
    limits: list[LimitRuleInput] | None = None,
) -> tuple[str, str]:
    async with SessionLocal() as session:
        service = ApiKeysService(ApiKeysRepository(session))
        created = await service.create_key(
            ApiKeyCreateData(
                name=name,
                allowed_models=None,
                limits=limits or [],
            )
        )
    return created.id, created.key


async def _seed_upstream_usage(*, now) -> None:
    suffix = str(int(now.timestamp() * 1_000_000))
    account_a_id = f"acc-plus-a-{suffix}"
    account_b_id = f"acc-plus-b-{suffix}"

    async with SessionLocal() as session:
        session.add_all(
            [
                Account(
                    id=account_a_id,
                    chatgpt_account_id=f"chatgpt-plus-a-{suffix}",
                    email=f"plus-a-{suffix}@example.com",
                    plan_type="plus",
                    access_token_encrypted=b"a",
                    refresh_token_encrypted=b"b",
                    id_token_encrypted=b"c",
                    last_refresh=now,
                    status=AccountStatus.ACTIVE,
                    reset_at=None,
                ),
                Account(
                    id=account_b_id,
                    chatgpt_account_id=f"chatgpt-plus-b-{suffix}",
                    email=f"plus-b-{suffix}@example.com",
                    plan_type="plus",
                    access_token_encrypted=b"d",
                    refresh_token_encrypted=b"e",
                    id_token_encrypted=b"f",
                    last_refresh=now,
                    status=AccountStatus.ACTIVE,
                    reset_at=None,
                ),
                UsageHistory(
                    account_id=account_a_id,
                    recorded_at=now,
                    window="primary",
                    used_percent=10.0,
                    reset_at=int((now + timedelta(hours=4)).timestamp()),
                    window_minutes=300,
                ),
                UsageHistory(
                    account_id=account_b_id,
                    recorded_at=now,
                    window="primary",
                    used_percent=20.0,
                    reset_at=int((now + timedelta(hours=4)).timestamp()),
                    window_minutes=300,
                ),
                UsageHistory(
                    account_id=account_a_id,
                    recorded_at=now,
                    window="secondary",
                    used_percent=20.0,
                    reset_at=int((now + timedelta(days=6)).timestamp()),
                    window_minutes=10080,
                ),
                UsageHistory(
                    account_id=account_b_id,
                    recorded_at=now,
                    window="secondary",
                    used_percent=30.0,
                    reset_at=int((now + timedelta(days=6)).timestamp()),
                    window_minutes=10080,
                ),
            ]
        )
        await session.commit()


async def _seed_upstream_usage_partial(
    *,
    now,
    windows: tuple[str, ...],
    primary_reset_at: int | None | object = ...,
    secondary_reset_at: int | None | object = ...,
) -> None:
    suffix = str(int(now.timestamp() * 1_000_000))
    account_id = f"acc-plus-partial-{suffix}"

    entries = [
        Account(
            id=account_id,
            chatgpt_account_id=f"chatgpt-plus-partial-{suffix}",
            email=f"plus-partial-{suffix}@example.com",
            plan_type="plus",
            access_token_encrypted=b"a",
            refresh_token_encrypted=b"b",
            id_token_encrypted=b"c",
            last_refresh=now,
            status=AccountStatus.ACTIVE,
            reset_at=None,
        )
    ]

    if "primary" in windows:
        if primary_reset_at is ...:
            primary_reset: int | None = int((now + timedelta(hours=4)).timestamp())
        else:
            primary_reset = primary_reset_at
        entries.append(
            UsageHistory(
                account_id=account_id,
                recorded_at=now,
                window="primary",
                used_percent=10.0,
                reset_at=primary_reset,
                window_minutes=300,
            )
        )
    if "secondary" in windows:
        if secondary_reset_at is ...:
            secondary_reset: int | None = int((now + timedelta(days=6)).timestamp())
        else:
            secondary_reset = secondary_reset_at
        entries.append(
            UsageHistory(
                account_id=account_id,
                recorded_at=now,
                window="secondary",
                used_percent=20.0,
                reset_at=secondary_reset,
                window_minutes=10080,
            )
        )

    async with SessionLocal() as session:
        session.add_all(entries)
        await session.commit()


async def _seed_upstream_usage_with_statuses(*, now) -> None:
    suffix = str(int(now.timestamp() * 1_000_000))
    active_id = f"acc-plus-active-{suffix}"
    paused_id = f"acc-plus-paused-{suffix}"
    deactivated_id = f"acc-plus-deactivated-{suffix}"

    async with SessionLocal() as session:
        session.add_all(
            [
                Account(
                    id=active_id,
                    chatgpt_account_id=f"chatgpt-plus-active-{suffix}",
                    email=f"plus-active-{suffix}@example.com",
                    plan_type="plus",
                    access_token_encrypted=b"a",
                    refresh_token_encrypted=b"b",
                    id_token_encrypted=b"c",
                    last_refresh=now,
                    status=AccountStatus.ACTIVE,
                    reset_at=None,
                ),
                Account(
                    id=paused_id,
                    chatgpt_account_id=f"chatgpt-plus-paused-{suffix}",
                    email=f"plus-paused-{suffix}@example.com",
                    plan_type="plus",
                    access_token_encrypted=b"d",
                    refresh_token_encrypted=b"e",
                    id_token_encrypted=b"f",
                    last_refresh=now,
                    status=AccountStatus.PAUSED,
                    reset_at=None,
                ),
                Account(
                    id=deactivated_id,
                    chatgpt_account_id=f"chatgpt-plus-deactivated-{suffix}",
                    email=f"plus-deactivated-{suffix}@example.com",
                    plan_type="plus",
                    access_token_encrypted=b"g",
                    refresh_token_encrypted=b"h",
                    id_token_encrypted=b"i",
                    last_refresh=now,
                    status=AccountStatus.DEACTIVATED,
                    reset_at=None,
                ),
                UsageHistory(
                    account_id=active_id,
                    recorded_at=now,
                    window="primary",
                    used_percent=20.0,
                    reset_at=int((now + timedelta(hours=4)).timestamp()),
                    window_minutes=300,
                ),
                UsageHistory(
                    account_id=paused_id,
                    recorded_at=now,
                    window="primary",
                    used_percent=100.0,
                    reset_at=int((now + timedelta(hours=4)).timestamp()),
                    window_minutes=300,
                ),
                UsageHistory(
                    account_id=deactivated_id,
                    recorded_at=now,
                    window="primary",
                    used_percent=60.0,
                    reset_at=int((now + timedelta(hours=4)).timestamp()),
                    window_minutes=300,
                ),
                UsageHistory(
                    account_id=active_id,
                    recorded_at=now,
                    window="secondary",
                    used_percent=25.0,
                    reset_at=int((now + timedelta(days=6)).timestamp()),
                    window_minutes=10080,
                ),
                UsageHistory(
                    account_id=paused_id,
                    recorded_at=now,
                    window="secondary",
                    used_percent=100.0,
                    reset_at=int((now + timedelta(days=6)).timestamp()),
                    window_minutes=10080,
                ),
                UsageHistory(
                    account_id=deactivated_id,
                    recorded_at=now,
                    window="secondary",
                    used_percent=75.0,
                    reset_at=int((now + timedelta(days=6)).timestamp()),
                    window_minutes=10080,
                ),
            ]
        )
        await session.commit()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("headers", "expected_message"),
    [
        ({}, "Missing API key in Authorization header"),
        ({"Authorization": "Bearer invalid-key"}, "Invalid API key"),
    ],
)
async def test_v1_usage_requires_valid_bearer_key_when_global_auth_disabled(async_client, headers, expected_message):
    response = await async_client.get("/v1/usage", headers=headers)

    assert response.status_code == 401
    payload = response.json()
    assert payload["error"]["code"] == "invalid_api_key"
    assert payload["error"]["message"] == expected_message


@pytest.mark.asyncio
async def test_v1_usage_returns_zero_usage_for_key_without_logs(async_client):
    _, plain_key = await _create_api_key(name="zero-usage")

    response = await async_client.get("/v1/usage", headers={"Authorization": f"Bearer {plain_key}"})

    assert response.status_code == 200
    assert response.json() == {
        "request_count": 0,
        "total_tokens": 0,
        "cached_input_tokens": 0,
        "total_cost_usd": 0.0,
        "limits": [],
    }


@pytest.mark.asyncio
async def test_v1_usage_scopes_usage_and_limits_to_authenticated_key(async_client):
    key_a_id, key_a = await _create_api_key(
        name="usage-key-a",
        limits=[LimitRuleInput(limit_type="total_tokens", limit_window="weekly", max_value=1_000)],
    )
    key_b_id, _ = await _create_api_key(
        name="usage-key-b",
        limits=[LimitRuleInput(limit_type="cost_usd", limit_window="monthly", max_value=5_000_000)],
    )

    now = utcnow()
    limit_a_total_reset = now + timedelta(days=6)
    limit_a_cost_reset = now + timedelta(days=20)
    limit_b_reset = now + timedelta(days=20)

    async with SessionLocal() as session:
        repo = ApiKeysRepository(session)
        await repo.replace_limits(
            key_a_id,
            [
                ApiKeyLimit(
                    api_key_id=key_a_id,
                    limit_type=LimitType.TOTAL_TOKENS,
                    limit_window=LimitWindow.WEEKLY,
                    max_value=1_000,
                    current_value=420,
                    model_filter="gpt-5.4",
                    reset_at=limit_a_total_reset,
                ),
                ApiKeyLimit(
                    api_key_id=key_a_id,
                    limit_type=LimitType.COST_USD,
                    limit_window=LimitWindow.MONTHLY,
                    max_value=10_000_000,
                    current_value=2_500_000,
                    model_filter=None,
                    reset_at=limit_a_cost_reset,
                ),
            ],
        )
        await repo.replace_limits(
            key_b_id,
            [
                ApiKeyLimit(
                    api_key_id=key_b_id,
                    limit_type=LimitType.COST_USD,
                    limit_window=LimitWindow.MONTHLY,
                    max_value=5_000_000,
                    current_value=5,
                    model_filter=None,
                    reset_at=limit_b_reset,
                )
            ],
        )

        logs = RequestLogsRepository(session)
        await logs.add_log(
            account_id=None,
            api_key_id=key_a_id,
            request_id="req_v1_usage_a1",
            model="gpt-5.4",
            input_tokens=100,
            output_tokens=25,
            cached_input_tokens=20,
            latency_ms=100,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=2),
        )
        await logs.add_log(
            account_id=None,
            api_key_id=key_a_id,
            request_id="req_v1_usage_a2",
            model="gpt-5.4",
            input_tokens=10,
            output_tokens=5,
            cached_input_tokens=2,
            latency_ms=80,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=1),
        )
        await logs.add_log(
            account_id=None,
            api_key_id=key_b_id,
            request_id="req_v1_usage_b1",
            model="gpt-5.4-mini",
            input_tokens=999,
            output_tokens=111,
            cached_input_tokens=50,
            latency_ms=90,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=3),
        )

    response = await async_client.get("/v1/usage", headers={"Authorization": f"Bearer {key_a}"})

    assert response.status_code == 200
    payload = response.json()
    assert payload["request_count"] == 2
    assert payload["total_tokens"] == 140
    assert payload["cached_input_tokens"] == 22
    assert payload["total_cost_usd"] > 0
    assert payload["limits"] == [
        {
            "limit_type": "total_tokens",
            "limit_window": "weekly",
            "max_value": 1000,
            "current_value": 420,
            "remaining_value": 580,
            "model_filter": "gpt-5.4",
            "reset_at": limit_a_total_reset.isoformat() + "Z",
            "source": "api_key_limit",
        },
        {
            "limit_type": "cost_usd",
            "limit_window": "monthly",
            "max_value": 10000000,
            "current_value": 2500000,
            "remaining_value": 7500000,
            "model_filter": None,
            "reset_at": limit_a_cost_reset.isoformat() + "Z",
            "source": "api_key_limit",
        },
    ]


@pytest.mark.asyncio
async def test_v1_usage_still_works_when_global_api_key_auth_is_disabled(async_client):
    _, plain_key = await _create_api_key(name="self-usage-auth-disabled")

    response = await async_client.get("/v1/usage", headers={"Authorization": f"Bearer {plain_key}"})

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_v1_usage_returns_aggregate_credit_limits_when_upstream_usage_exists(async_client):
    _, plain_key = await _create_api_key(name="fallback-aggregate")
    now = utcnow()
    await _seed_upstream_usage(now=now)

    response = await async_client.get("/v1/usage", headers={"Authorization": f"Bearer {plain_key}"})

    assert response.status_code == 200
    payload = response.json()
    assert payload["limits"][0] == {
        "limit_type": "credits",
        "limit_window": "5h",
        "max_value": 450,
        "current_value": 68,
        "remaining_value": 382,
        "model_filter": None,
        "reset_at": payload["limits"][0]["reset_at"],
        "source": "aggregate",
    }
    assert payload["limits"][1] == {
        "limit_type": "credits",
        "limit_window": "7d",
        "max_value": 15120,
        "current_value": 3780,
        "remaining_value": 11340,
        "model_filter": None,
        "reset_at": payload["limits"][1]["reset_at"],
        "source": "aggregate",
    }
    assert payload["limits"][0]["reset_at"].endswith("Z")
    assert payload["limits"][1]["reset_at"].endswith("Z")


@pytest.mark.asyncio
async def test_v1_usage_overrides_aggregate_credit_windows_with_api_key_credit_limits(async_client):
    key_id, plain_key = await _create_api_key(
        name="credit-override",
        limits=[
            LimitRuleInput(limit_type="credits", limit_window="5h", max_value=60),
            LimitRuleInput(limit_type="credits", limit_window="7d", max_value=1000),
        ],
    )
    now = utcnow()
    await _seed_upstream_usage(now=now)

    async with SessionLocal() as session:
        repo = ApiKeysRepository(session)
        await repo.replace_limits(
            key_id,
            [
                ApiKeyLimit(
                    api_key_id=key_id,
                    limit_type=LimitType.CREDITS,
                    limit_window=LimitWindow.FIVE_HOURS,
                    max_value=60,
                    current_value=999,
                    model_filter=None,
                    reset_at=now + timedelta(hours=5),
                ),
                ApiKeyLimit(
                    api_key_id=key_id,
                    limit_type=LimitType.CREDITS,
                    limit_window=LimitWindow.SEVEN_DAYS,
                    max_value=1000,
                    current_value=10,
                    model_filter=None,
                    reset_at=now + timedelta(days=7),
                ),
            ],
        )

    response = await async_client.get("/v1/usage", headers={"Authorization": f"Bearer {plain_key}"})

    assert response.status_code == 200
    payload = response.json()
    assert payload["limits"] == [
        {
            "limit_type": "credits",
            "limit_window": "5h",
            "max_value": 60,
            "current_value": 60,
            "remaining_value": 0,
            "model_filter": None,
            "reset_at": payload["limits"][0]["reset_at"],
            "source": "api_key_override",
        },
        {
            "limit_type": "credits",
            "limit_window": "7d",
            "max_value": 1000,
            "current_value": 1000,
            "remaining_value": 0,
            "model_filter": None,
            "reset_at": payload["limits"][1]["reset_at"],
            "source": "api_key_override",
        },
    ]


@pytest.mark.asyncio
async def test_v1_usage_prefers_raw_limits_when_aggregate_credit_pair_is_partial(async_client):
    key_id, plain_key = await _create_api_key(
        name="fallback-partial-raw",
        limits=[
            LimitRuleInput(limit_type="total_tokens", limit_window="daily", max_value=300),
            LimitRuleInput(limit_type="total_tokens", limit_window="weekly", max_value=1000),
        ],
    )
    now = utcnow()
    await _seed_upstream_usage_partial(now=now, windows=("primary",))

    daily_reset = now + timedelta(hours=2)
    weekly_reset = now + timedelta(days=4)

    async with SessionLocal() as session:
        repo = ApiKeysRepository(session)
        await repo.replace_limits(
            key_id,
            [
                ApiKeyLimit(
                    api_key_id=key_id,
                    limit_type=LimitType.TOTAL_TOKENS,
                    limit_window=LimitWindow.DAILY,
                    max_value=300,
                    current_value=50,
                    model_filter=None,
                    reset_at=daily_reset,
                ),
                ApiKeyLimit(
                    api_key_id=key_id,
                    limit_type=LimitType.TOTAL_TOKENS,
                    limit_window=LimitWindow.WEEKLY,
                    max_value=1000,
                    current_value=200,
                    model_filter=None,
                    reset_at=weekly_reset,
                ),
            ],
        )

    response = await async_client.get("/v1/usage", headers={"Authorization": f"Bearer {plain_key}"})

    assert response.status_code == 200
    payload = response.json()
    assert payload["limits"] == [
        {
            "limit_type": "total_tokens",
            "limit_window": "daily",
            "max_value": 300,
            "current_value": 50,
            "remaining_value": 250,
            "model_filter": None,
            "reset_at": daily_reset.isoformat() + "Z",
            "source": "api_key_limit",
        },
        {
            "limit_type": "total_tokens",
            "limit_window": "weekly",
            "max_value": 1000,
            "current_value": 200,
            "remaining_value": 800,
            "model_filter": None,
            "reset_at": weekly_reset.isoformat() + "Z",
            "source": "api_key_limit",
        },
    ]


@pytest.mark.asyncio
async def test_v1_usage_falls_back_to_raw_credit_limits_when_aggregate_reset_is_missing(async_client):
    key_id, plain_key = await _create_api_key(
        name="fallback-missing-reset",
        limits=[
            LimitRuleInput(limit_type="credits", limit_window="5h", max_value=60),
            LimitRuleInput(limit_type="credits", limit_window="7d", max_value=1000),
        ],
    )
    now = utcnow()
    await _seed_upstream_usage_partial(now=now, windows=("primary", "secondary"), primary_reset_at=None)

    primary_reset = now + timedelta(hours=5)
    secondary_reset = now + timedelta(days=7)

    async with SessionLocal() as session:
        repo = ApiKeysRepository(session)
        await repo.replace_limits(
            key_id,
            [
                ApiKeyLimit(
                    api_key_id=key_id,
                    limit_type=LimitType.CREDITS,
                    limit_window=LimitWindow.FIVE_HOURS,
                    max_value=60,
                    current_value=12,
                    model_filter=None,
                    reset_at=primary_reset,
                ),
                ApiKeyLimit(
                    api_key_id=key_id,
                    limit_type=LimitType.CREDITS,
                    limit_window=LimitWindow.SEVEN_DAYS,
                    max_value=1000,
                    current_value=250,
                    model_filter=None,
                    reset_at=secondary_reset,
                ),
            ],
        )
        await session.commit()

    response = await async_client.get("/v1/usage", headers={"Authorization": f"Bearer {plain_key}"})

    assert response.status_code == 200
    assert response.json()["limits"] == [
        {
            "limit_type": "credits",
            "limit_window": "5h",
            "max_value": 60,
            "current_value": 12,
            "remaining_value": 48,
            "model_filter": None,
            "reset_at": primary_reset.isoformat() + "Z",
            "source": "api_key_override",
        },
        {
            "limit_type": "credits",
            "limit_window": "7d",
            "max_value": 1000,
            "current_value": 250,
            "remaining_value": 750,
            "model_filter": None,
            "reset_at": secondary_reset.isoformat() + "Z",
            "source": "api_key_override",
        },
    ]


@pytest.mark.asyncio
async def test_v1_usage_ignores_paused_and_deactivated_accounts_in_aggregate_credit_windows(async_client):
    _, plain_key = await _create_api_key(name="aggregate-active-only")
    now = utcnow()
    await _seed_upstream_usage_with_statuses(now=now)

    response = await async_client.get("/v1/usage", headers={"Authorization": f"Bearer {plain_key}"})

    assert response.status_code == 200
    payload = response.json()
    assert payload["limits"] == [
        {
            "limit_type": "credits",
            "limit_window": "5h",
            "max_value": 225,
            "current_value": 45,
            "remaining_value": 180,
            "model_filter": None,
            "reset_at": payload["limits"][0]["reset_at"],
            "source": "aggregate",
        },
        {
            "limit_type": "credits",
            "limit_window": "7d",
            "max_value": 7560,
            "current_value": 1890,
            "remaining_value": 5670,
            "model_filter": None,
            "reset_at": payload["limits"][1]["reset_at"],
            "source": "aggregate",
        },
    ]



"""add performance indexes for sqlite startup query paths

Revision ID: 20260308_000000_add_sqlite_performance_indexes
Revises: 20260307_000000_add_api_key_enforcement_fields
Create Date: 2026-03-08
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "20260308_000000_add_sqlite_performance_indexes"
down_revision = "20260307_000000_add_api_key_enforcement_fields"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        sa.text(
            """
            CREATE INDEX IF NOT EXISTS idx_usage_window_account_latest
            ON usage_history (coalesce("window", 'primary'), account_id, recorded_at DESC, id DESC)
            """
        )
    )
    op.execute(
        sa.text(
            """
            CREATE INDEX IF NOT EXISTS idx_logs_requested_at_id
            ON request_logs (requested_at DESC, id DESC)
            """
        )
    )


def downgrade() -> None:
    op.execute(sa.text("DROP INDEX IF EXISTS idx_logs_requested_at_id"))
    op.execute(sa.text("DROP INDEX IF EXISTS idx_usage_window_account_latest"))



from __future__ import annotations

from collections.abc import Sequence

from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from app.db.models import ApiFirewallAllowlist


class FirewallRepositoryConflictError(ValueError):
    pass


class FirewallRepository:
    def __init__(self, session: AsyncSession) -> None:
        self._session = session

    async def list_entries(self) -> Sequence[ApiFirewallAllowlist]:
        result = await self._session.execute(
            select(ApiFirewallAllowlist).order_by(ApiFirewallAllowlist.created_at, ApiFirewallAllowlist.ip_address)
        )
        return list(result.scalars().all())

    async def list_ip_addresses(self) -> set[str]:
        result = await self._session.execute(select(ApiFirewallAllowlist.ip_address))
        return {row[0] for row in result.all()}

    async def exists(self, ip_address: str) -> bool:
        row = await self._session.get(ApiFirewallAllowlist, ip_address)
        return row is not None

    async def add(self, ip_address: str) -> ApiFirewallAllowlist:
        row = ApiFirewallAllowlist(ip_address=ip_address)
        self._session.add(row)
        try:
            await self._session.commit()
        except IntegrityError as exc:
            await self._session.rollback()
            raise FirewallRepositoryConflictError("IP address already exists") from exc
        await self._session.refresh(row)
        return row

    async def delete(self, ip_address: str) -> bool:
        row = await self._session.get(ApiFirewallAllowlist, ip_address)
        if row is None:
            return False
        await self._session.delete(row)
        await self._session.commit()
        return True



#!/usr/bin/env python3
"""PreToolUse hook — enforce guardrail skills before Edit/Write operations."""

import json
import os
import re
import sys
from fnmatch import fnmatch
from pathlib import Path

try:
    from _analytics import emit_event
except ImportError:

    def emit_event(*_a: object, **_k: object) -> None:
        pass


def load_skill_rules() -> dict:
    """Load skill-rules.json relative to this hook's location."""
    hook_dir = Path(__file__).resolve().parent
    rules_path = hook_dir.parent / "skills" / "skill-rules.json"
    with open(rules_path) as f:
        return json.load(f)


def load_session_state(session_id: str) -> dict:
    """Load session state for tracking which skills have been used."""
    hook_dir = Path(__file__).resolve().parent
    state_path = hook_dir / "state" / f"skills-used-{session_id}.json"
    if state_path.exists():
        try:
            with open(state_path) as f:
                return json.load(f)
        except (json.JSONDecodeError, ValueError):
            pass
    return {"suggestedSkills": [], "usedSkills": []}


def match_path_patterns(file_path: str, patterns: list[str]) -> bool:
    """Check if file_path matches any glob pattern."""
    return any(fnmatch(file_path, pat) for pat in patterns)


def match_content_patterns(file_path: str, patterns: list[str]) -> bool:
    """Check if file content matches any regex pattern."""
    try:
        content = Path(file_path).read_text(errors="ignore")
        return any(re.search(pat, content) for pat in patterns)
    except (FileNotFoundError, PermissionError):
        return False


def check_pass_state(pass_state_file: str) -> bool:
    """Check if a pass state file exists and has result=PASS."""
    hook_dir = Path(__file__).resolve().parent
    state_path = hook_dir / "state" / pass_state_file
    if not state_path.exists():
        return False
    try:
        data = json.loads(state_path.read_text())
        return data.get("result") == "PASS"
    except (json.JSONDecodeError, PermissionError):
        return False


def check_file_markers(file_path: str, markers: list[str]) -> bool:
    """Check if file contains any skip markers."""
    try:
        content = Path(file_path).read_text(errors="ignore")
        return any(marker in content for marker in markers)
    except (FileNotFoundError, PermissionError):
        return False


def main() -> None:
    try:
        input_data = json.loads(sys.stdin.read())
    except (json.JSONDecodeError, EOFError):
        sys.exit(0)  # fail-open

    session_id = input_data.get("session_id", "unknown")
    tool_input = input_data.get("tool_input", {})
    file_path = tool_input.get("file_path", "")

    if not file_path:
        sys.exit(0)

    try:
        rules = load_skill_rules()
    except (FileNotFoundError, json.JSONDecodeError):
        sys.exit(0)  # fail-open

    skills = rules.get("skills", {})

    session_state = load_session_state(session_id)

    # --- Phase 1: Hard block guardrails ---
    guardrails = {
        name: rule
        for name, rule in skills.items()
        if rule.get("type") == "guardrail" and rule.get("enforcement") == "block"
    }

    for name, rule in guardrails.items():
        file_triggers = rule.get("fileTriggers")
        if not file_triggers:
            continue

        path_patterns = file_triggers.get("pathPatterns", [])
        if not match_path_patterns(file_path, path_patterns):
            continue

        path_exclusions = file_triggers.get("pathExclusions", [])
        if path_exclusions and match_path_patterns(file_path, path_exclusions):
            continue

        content_patterns = file_triggers.get("contentPatterns", [])
        if content_patterns and not match_content_patterns(file_path, content_patterns):
            continue

        # --- Skip conditions ---
        skip = rule.get("skipConditions", {})

        pass_state_file = skip.get("passStateFile")
        if pass_state_file and check_pass_state(pass_state_file):
            continue

        if skip.get("sessionSkillUsed") and name in session_state.get("usedSkills", []):
            continue

        file_markers = skip.get("fileMarkers", [])
        if file_markers and check_file_markers(file_path, file_markers):
            continue

        env_override = skip.get("envOverride")
        if env_override and os.environ.get(env_override):
            continue

        # All checks passed — block
        emit_event(
            session_id,
            "hook.invoked",
            {
                "hook": "skill_guard",
                "trigger": "PreToolUse",
                "outcome": "blocked",
                "matched_count": 1,
                "exit_code": 2,
            },
        )
        block_message = rule.get(
            "blockMessage",
            f"BLOCKED: Skill '{name}' must be invoked before editing this file.\nUse Skill tool: '{name}'",
        )
        print(block_message, file=sys.stderr)
        sys.exit(2)

    # --- Phase 2: Remind enforcement (block until skill invoked in session) ---
    remind_rules = {
        name: rule for name, rule in skills.items() if rule.get("enforcement") == "remind" and rule.get("fileTriggers")
    }

    for name, rule in remind_rules.items():
        file_triggers = rule.get("fileTriggers", {})

        path_patterns = file_triggers.get("pathPatterns", [])
        if not match_path_patterns(file_path, path_patterns):
            continue

        path_exclusions = file_triggers.get("pathExclusions", [])
        if path_exclusions and match_path_patterns(file_path, path_exclusions):
            continue

        content_patterns = file_triggers.get("contentPatterns", [])
        if content_patterns and not match_content_patterns(file_path, content_patterns):
            continue

        # --- Skip conditions ---
        skip = rule.get("skipConditions", {})

        if skip.get("sessionSkillUsed") and name in session_state.get("usedSkills", []):
            continue

        env_override = skip.get("envOverride")
        if env_override and os.environ.get(env_override):
            continue

        # Block with convention reminder
        emit_event(
            session_id,
            "hook.invoked",
            {
                "hook": "skill_guard",
                "trigger": "PreToolUse",
                "outcome": "remind_blocked",
                "matched_count": 1,
                "exit_code": 2,
            },
        )
        block_message = rule.get(
            "blockMessage",
            f"BLOCKED: Run /{name} first.\n"
            f"You must invoke this skill before editing this file.\n\n"
            f"→ Skill tool: '{name}'",
        )
        print(block_message, file=sys.stderr)
        sys.exit(2)

    emit_event(
        session_id,
        "hook.invoked",
        {
            "hook": "skill_guard",
            "trigger": "PreToolUse",
            "outcome": "passed",
            "matched_count": 0,
            "exit_code": 0,
        },
    )
    sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations

import base64
import json
from datetime import timezone

import pytest

import app.modules.proxy.provider_adapters as proxy_transport_module
from app.core.utils.time import utcnow
from app.db.session import SessionLocal
from app.modules.usage.repository import UsageRepository

pytestmark = pytest.mark.integration


def _encode_jwt(payload: dict) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


def _make_auth_json(account_id: str, email: str) -> dict:
    payload = {
        "email": email,
        "chatgpt_account_id": account_id,
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    return {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access-token",
            "refreshToken": "refresh-token",
            "accountId": account_id,
        },
    }


async def _import_account(async_client, account_id: str, email: str) -> str:
    auth_json = _make_auth_json(account_id, email)
    files = {"auth_json": ("auth.json", json.dumps(auth_json), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200
    return response.json()["accountId"]


def test_mini_and_large_requests_use_different_cache_keys():
    """Verify that gpt-5.4-mini and gpt-5.3-codex requests produce different prompt_cache_keys
    due to model-class prefix separation."""
    from app.core.openai.requests import ResponsesRequest
    from app.modules.proxy.service import _derive_prompt_cache_key

    mini_payload = ResponsesRequest(
        model="gpt-5.4-mini",
        instructions="You are helpful.",
        input=[{"role": "user", "content": "Hello"}],
    )

    large_payload = ResponsesRequest(
        model="gpt-5.3-codex",
        instructions="You are helpful.",
        input=[{"role": "user", "content": "Hello"}],
    )

    mini_key = _derive_prompt_cache_key(mini_payload, None)
    large_key = _derive_prompt_cache_key(large_payload, None)

    assert mini_key != large_key, "Mini and large models should produce different cache keys"
    assert mini_key.startswith("mini-"), f"Mini key should start with 'mini-', got {mini_key}"
    assert large_key.startswith("codex-"), f"Large key should start with 'codex-', got {large_key}"


def test_same_model_class_produces_same_cache_key():
    """Verify that two requests with the same model class produce the same cache key."""
    from app.core.openai.requests import ResponsesRequest
    from app.modules.proxy.service import _derive_prompt_cache_key

    payload = ResponsesRequest(
        model="gpt-5.3-codex",
        instructions="You are helpful.",
        input=[{"role": "user", "content": "Hello"}],
    )

    key1 = _derive_prompt_cache_key(payload, None)
    key2 = _derive_prompt_cache_key(payload, None)

    assert key1 == key2, "Same model class should produce identical cache keys"


def test_prompt_cache_bridge_idle_ttl_from_settings():
    """Verify that PROMPT_CACHE bridge idle TTL is read from dashboard settings."""
    from app.db.models import StickySessionKind
    from app.modules.proxy.service import _AffinityPolicy, _effective_http_bridge_idle_ttl_seconds

    affinity = _AffinityPolicy(
        key="cache-key-123",
        kind=StickySessionKind.PROMPT_CACHE,
    )

    result = _effective_http_bridge_idle_ttl_seconds(
        affinity=affinity,
        idle_ttl_seconds=120.0,
        codex_idle_ttl_seconds=900.0,
        prompt_cache_idle_ttl_seconds=3600.0,
    )

    assert result == 3600.0, "PROMPT_CACHE should use prompt_cache_idle_ttl_seconds"


def test_codex_session_idle_ttl_unchanged():
    """Verify that CODEX_SESSION idle TTL behavior is unchanged."""
    from app.db.models import StickySessionKind
    from app.modules.proxy.service import _AffinityPolicy, _effective_http_bridge_idle_ttl_seconds

    affinity = _AffinityPolicy(
        key="session-123",
        kind=StickySessionKind.CODEX_SESSION,
    )

    result = _effective_http_bridge_idle_ttl_seconds(
        affinity=affinity,
        idle_ttl_seconds=120.0,
        codex_idle_ttl_seconds=900.0,
        prompt_cache_idle_ttl_seconds=3600.0,
    )

    assert result == 900.0, "CODEX_SESSION should use max(idle_ttl_seconds, codex_idle_ttl_seconds)"


def test_model_class_extraction_for_all_model_types():
    """Verify model class extraction works correctly for mini, codex, and standard models."""
    from app.modules.proxy.service import _extract_model_class

    assert _extract_model_class("gpt-5.4-mini") == "mini"
    assert _extract_model_class("gpt-4-mini") == "mini"

    assert _extract_model_class("gpt-5.3-codex") == "codex"
    assert _extract_model_class("gpt-5.3-codex-spark") == "codex"
    assert _extract_model_class("gpt-5.1-codex-mini") == "codex"

    # Standard models (default)
    assert _extract_model_class("gpt-5.4") == "std"
    assert _extract_model_class("gpt-4") == "std"
    assert _extract_model_class("gpt-4-turbo") == "std"


@pytest.mark.asyncio
async def test_prompt_cache_reallocates_when_usage_exceeds_configured_budget_threshold(async_client, monkeypatch):
    settings_response = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "stickyReallocationBudgetThresholdPct": 80.0,
            "preferEarlierResetAccounts": False,
            "routingStrategy": "usage_weighted",
        },
    )
    assert settings_response.status_code == 200

    acc_a_id = await _import_account(async_client, "acc_budget_a", "budget_a@example.com")
    acc_b_id = await _import_account(async_client, "acc_budget_b", "budget_b@example.com")

    now_epoch = int(utcnow().replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=20.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    seen: list[str] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kwargs):
        seen.append(account_id)
        yield 'data: {"type":"response.completed","response":{"id":"resp_budget"}}\n\n'

    monkeypatch.setattr(proxy_transport_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.1",
        "instructions": "hi",
        "input": [],
        "stream": True,
        "prompt_cache_key": "budget-threshold-key",
    }

    first = await async_client.post("/backend-api/codex/responses", json=payload)
    assert first.status_code == 200
    assert seen == ["acc_budget_a"]

    async with SessionLocal() as session:
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=acc_a_id,
            used_percent=85.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=acc_b_id,
            used_percent=5.0,
            window="primary",
            reset_at=now_epoch + 3600,
            window_minutes=300,
        )

    second = await async_client.post("/backend-api/codex/responses", json=payload)
    assert second.status_code == 200
    assert seen == ["acc_budget_a", "acc_budget_b"]


def test_owner_mismatch_raises_409_for_retry() -> None:
    """Verify that the bridge mismatch block raises ProxyResponseError(409).

    When a request lands on a non-owner replica, the mismatch block must raise
    a 409 so the client retries instead of silently creating a duplicate local
    bridge that breaks turn-state continuity.
    """
    import inspect

    from app.modules.proxy import service as proxy_service_module

    source = inspect.getsource(proxy_service_module)
    assert "owner_mismatch_retry" in source, (
        "Expected 'owner_mismatch_retry' log event in service.py. "
        "The mismatch block should raise 409 for retry, not fall through."
    )



"""add index on request_logs.requested_at

Revision ID: 20260215_000000_add_idx_logs_requested_at
Revises: 20260214_000000_add_api_key_limits
Create Date: 2026-02-15
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260215_000000_add_idx_logs_requested_at"
down_revision = "20260214_000000_add_api_key_limits"
branch_labels = None
depends_on = None


def _index_exists(connection: Connection, index_name: str, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return False
    indexes = inspector.get_indexes(table_name)
    return any(idx["name"] == index_name for idx in indexes)


def upgrade() -> None:
    bind = op.get_bind()
    if not _index_exists(bind, "idx_logs_requested_at", "request_logs"):
        op.create_index("idx_logs_requested_at", "request_logs", ["requested_at"])


def downgrade() -> None:
    bind = op.get_bind()
    if _index_exists(bind, "idx_logs_requested_at", "request_logs"):
        op.drop_index("idx_logs_requested_at", table_name="request_logs")



"""add bridge gateway-safe mode to dashboard settings

Revision ID: 20260407_000000_add_bridge_gateway_safe_mode
Revises: 20260403_000000_add_credit_api_key_limit_values
Create Date: 2026-04-07
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

revision = "20260407_000000_add_bridge_gateway_safe_mode"
down_revision = "20260403_000000_add_credit_api_key_limit_values"
branch_labels = None
depends_on = None


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {column["name"] for column in inspector.get_columns(table_name)}


def upgrade() -> None:
    bind = op.get_bind()
    columns = _columns(bind, "dashboard_settings")
    if not columns or "http_responses_session_bridge_gateway_safe_mode" in columns:
        return

    with op.batch_alter_table("dashboard_settings") as batch_op:
        batch_op.add_column(
            sa.Column(
                "http_responses_session_bridge_gateway_safe_mode",
                sa.Boolean(),
                nullable=False,
                server_default=sa.false(),
            )
        )


def downgrade() -> None:
    bind = op.get_bind()
    columns = _columns(bind, "dashboard_settings")
    if not columns or "http_responses_session_bridge_gateway_safe_mode" not in columns:
        return

    with op.batch_alter_table("dashboard_settings") as batch_op:
        batch_op.drop_column("http_responses_session_bridge_gateway_safe_mode")



from __future__ import annotations

import time

from starlette.types import ASGIApp, Message, Receive, Scope, Send

from app.core.metrics.prometheus import (
    PROMETHEUS_AVAILABLE,
    active_connections,
    request_duration_seconds,
    requests_total,
)


def _normalize_path(path: str) -> str:
    if path.startswith("/v1/"):
        return "/v1/..."
    if path.startswith("/api/"):
        return "/api/..."
    if path.startswith("/health/"):
        return "/health/..."
    if len(path) > 50:
        return path[:50]
    return path or "/"


class MetricsMiddleware:
    def __init__(self, app: ASGIApp, *, enabled: bool = True) -> None:
        self.app = app
        self.enabled = enabled

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http" or not self.enabled or not PROMETHEUS_AVAILABLE:
            await self.app(scope, receive, send)
            return

        assert active_connections is not None
        assert requests_total is not None
        assert request_duration_seconds is not None

        start = time.monotonic()
        status_code = 500
        method = scope.get("method", "GET")
        path = _normalize_path(scope.get("path", "/"))

        active_connections.inc()

        async def send_wrapper(message: Message) -> None:
            nonlocal status_code
            if message["type"] == "http.response.start":
                status_code = int(message["status"])
            await send(message)

        try:
            await self.app(scope, receive, send_wrapper)
        finally:
            duration = time.monotonic() - start
            requests_total.labels(method=method, path=path, status=str(status_code)).inc()
            request_duration_seconds.labels(method=method, path=path).observe(duration)
            active_connections.dec()


__all__ = ["MetricsMiddleware", "_normalize_path"]



from __future__ import annotations

from sqlalchemy import or_, update
from sqlalchemy.ext.asyncio import AsyncSession

from app.db.models import DashboardSettings
from app.modules.settings.repository import SettingsRepository

_SETTINGS_ID = 1


class DashboardAuthRepository:
    def __init__(self, session: AsyncSession) -> None:
        self._session = session
        self._settings_repository = SettingsRepository(session)

    async def get_settings(self) -> DashboardSettings:
        return await self._settings_repository.get_or_create()

    async def set_totp_secret(self, secret_encrypted: bytes | None) -> DashboardSettings:
        row = await self._settings_repository.get_or_create()
        row.totp_secret_encrypted = secret_encrypted
        row.totp_last_verified_step = None
        if secret_encrypted is None:
            row.totp_required_on_login = False
        await self._settings_repository.commit_refresh(row)
        return row

    async def set_password_hash(self, password_hash: str) -> DashboardSettings:
        row = await self._settings_repository.get_or_create()
        row.password_hash = password_hash
        await self._settings_repository.commit_refresh(row)
        return row

    async def try_set_password_hash(self, password_hash: str) -> bool:
        await self._settings_repository.get_or_create()
        result = await self._session.execute(
            update(DashboardSettings)
            .where(DashboardSettings.id == _SETTINGS_ID)
            .where(DashboardSettings.password_hash.is_(None))
            .values(password_hash=password_hash)
            .returning(DashboardSettings.id)
        )
        await self._session.commit()
        return result.scalar_one_or_none() is not None

    async def get_password_hash(self) -> str | None:
        row = await self._settings_repository.get_or_create()
        return row.password_hash

    async def clear_password_and_totp(self) -> DashboardSettings:
        row = await self._settings_repository.get_or_create()
        row.password_hash = None
        row.totp_required_on_login = False
        row.totp_secret_encrypted = None
        row.totp_last_verified_step = None
        await self._settings_repository.commit_refresh(row)
        return row

    async def try_advance_totp_last_verified_step(self, step: int) -> bool:
        await self._settings_repository.get_or_create()
        result = await self._session.execute(
            update(DashboardSettings)
            .where(DashboardSettings.id == _SETTINGS_ID)
            .where(
                or_(
                    DashboardSettings.totp_last_verified_step.is_(None),
                    DashboardSettings.totp_last_verified_step < step,
                )
            )
            .values(totp_last_verified_step=step)
            .returning(DashboardSettings.id)
        )
        await self._session.commit()
        return result.scalar_one_or_none() is not None



from __future__ import annotations

from datetime import datetime

from pydantic import Field

from app.modules.shared.schemas import DashboardModel


class UsageWindow(DashboardModel):
    remaining_percent: float
    capacity_credits: float
    remaining_credits: float
    reset_at: datetime | None = None
    window_minutes: int | None = None


class UsageCost(DashboardModel):
    currency: str
    total_usd_7d: float = Field(alias="totalUsd7d")


class UsageMetrics(DashboardModel):
    requests_7d: int | None = Field(default=None, alias="requests7d")
    tokens_secondary_window: int | None = None
    cached_tokens_secondary_window: int | None = None
    error_rate_7d: float | None = Field(default=None, alias="errorRate7d")
    top_error: str | None = None


class UsageSummaryResponse(DashboardModel):
    primary_window: UsageWindow
    secondary_window: UsageWindow | None = None
    cost: UsageCost
    metrics: UsageMetrics | None = None


class UsageHistoryItem(DashboardModel):
    account_id: str
    remaining_percent_avg: float | None = None
    capacity_credits: float
    remaining_credits: float


class UsageHistoryResponse(DashboardModel):
    window_hours: int
    accounts: list[UsageHistoryItem] = Field(default_factory=list)


class UsageWindowResponse(DashboardModel):
    window_key: str
    window_minutes: int | None = None
    accounts: list[UsageHistoryItem] = Field(default_factory=list)


class TrendPoint(DashboardModel):
    t: datetime
    v: float


class MetricsTrends(DashboardModel):
    requests: list[TrendPoint] = Field(default_factory=list)
    tokens: list[TrendPoint] = Field(default_factory=list)
    cost: list[TrendPoint] = Field(default_factory=list)
    error_rate: list[TrendPoint] = Field(default_factory=list)



from __future__ import annotations

import sqlalchemy as sa
from alembic import op

revision = "20260328_120000_add_request_log_ttft"
down_revision = "20260328_000000_add_rate_limit_attempts"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    inspector = sa.inspect(bind)
    if not inspector.has_table("request_logs"):
        return

    columns = {column["name"] for column in inspector.get_columns("request_logs")}
    if "latency_first_token_ms" in columns:
        return

    with op.batch_alter_table("request_logs") as batch_op:
        batch_op.add_column(sa.Column("latency_first_token_ms", sa.Integer(), nullable=True))


def downgrade() -> None:
    bind = op.get_bind()
    inspector = sa.inspect(bind)
    if not inspector.has_table("request_logs"):
        return

    columns = {column["name"] for column in inspector.get_columns("request_logs")}
    if "latency_first_token_ms" not in columns:
        return

    with op.batch_alter_table("request_logs") as batch_op:
        batch_op.drop_column("latency_first_token_ms")
